# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This namespace contains various utility functions, classes, and type aliases.
"""


__all__ = (
    "Hook",
    "as_immutable",
    "mutable_copy",
    "Device",
    "DType",
    "DTypeAndDevice",
    "ErroneousResult",
    "RealOrVector",
    "Size",
    "SuppressSacredExperiment",
    "Vector",
    "as_tensor",
    "cast_tensors_in_container",
    "clip_tensor",
    "clone",
    "device_of",
    "device_of_container",
    "dtype_of",
    "dtype_of_container",
    "empty_tensor_like",
    "ensure_ray",
    "ensure_tensor_length_and_dtype",
    "expect_none",
    "is_bool",
    "is_bool_vector",
    "is_dtype_bool",
    "is_dtype_float",
    "is_dtype_integer",
    "is_dtype_object",
    "is_dtype_real",
    "is_integer",
    "is_integer_vector",
    "is_real",
    "is_real_vector",
    "is_sequence",
    "is_tensor_on_cpu",
    "make_empty",
    "make_gaussian",
    "make_I",
    "make_nan",
    "make_ones",
    "make_randint",
    "make_tensor",
    "make_uniform",
    "make_zeros",
    "modify_tensor",
    "multiply_rows_by_scalars",
    "numpy_copy",
    "rowwise_sum",
    "split_workload",
    "stdev_from_radius",
    "to_numpy_dtype",
    "to_stdev_init",
    "to_torch_dtype",
    "ObjectArray",
    "rank",
    "ReadOnlyTensor",
    "as_read_only_tensor",
    "read_only_tensor",
)


from . import hook, immutable, objectarray, ranking, readonlytensor, tensormaker
from .hook import Hook
from .immutable import as_immutable, mutable_copy
from .misc import (
    Device,
    DType,
    DTypeAndDevice,
    ErroneousResult,
    RealOrVector,
    Size,
    SuppressSacredExperiment,
    Vector,
    as_tensor,
    cast_tensors_in_container,
    clip_tensor,
    clone,
    device_of,
    device_of_container,
    dtype_of,
    dtype_of_container,
    empty_tensor_like,
    ensure_ray,
    ensure_tensor_length_and_dtype,
    expect_none,
    is_bool,
    is_bool_vector,
    is_dtype_bool,
    is_dtype_float,
    is_dtype_integer,
    is_dtype_object,
    is_dtype_real,
    is_integer,
    is_integer_vector,
    is_real,
    is_real_vector,
    is_sequence,
    is_tensor_on_cpu,
    make_empty,
    make_gaussian,
    make_I,
    make_nan,
    make_ones,
    make_randint,
    make_tensor,
    make_uniform,
    make_zeros,
    modify_tensor,
    multiply_rows_by_scalars,
    numpy_copy,
    rowwise_sum,
    split_workload,
    stdev_from_radius,
    to_numpy_dtype,
    to_stdev_init,
    to_torch_dtype,
)
from .objectarray import ObjectArray
from .ranking import rank
from .readonlytensor import ReadOnlyTensor, as_read_only_tensor, read_only_tensor



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evotorch.tools.hook import Hook


def test_hook():
    call_order = []

    def f():
        call_order.append("f")

    def g():
        call_order.append("g")

    my_hook = Hook([f, g])
    my_hook()

    assert call_order == ["f", "g"]


def test_hook_with_args():
    call_order = []

    def f(a, b, *, c):
        call_order.append("f")
        assert a == 1
        assert b == 2
        assert c == 3

    def g(a, b, *, c):
        call_order.append("g")
        assert a == 1
        assert b == 2
        assert c == 3

    my_hook = Hook([f, g])
    my_hook(1, 2, c=3)

    assert call_order == ["f", "g"]


def test_hook_with_dict_result():
    call_order = []

    def f():
        call_order.append("f")
        return {"a": 1}

    def h():
        call_order.append("h")
        return None

    def g():
        call_order.append("g")
        return {"b": 2}

    my_hook = Hook([f, h, g])
    result = my_hook()
    assert call_order == ["f", "h", "g"]
    assert set(result.keys()) == {"a", "b"}
    assert result["a"] == 1
    assert result["b"] == 2


def test_hook_with_list_result():
    call_order = []

    def f():
        call_order.append("f")
        return [1]

    def h():
        call_order.append("h")
        return None

    def g():
        call_order.append("g")
        return [2]

    my_hook = Hook([f, h, g])
    result = my_hook()
    assert call_order == ["f", "h", "g"]
    assert result == [1, 2]


def test_hook_with_default_args():
    call_order = []

    def f(a, b, *, c):
        call_order.append("f")
        assert a == 1
        assert b == 2
        assert c == 3

    def g(a, b, *, c):
        call_order.append("g")
        assert a == 1
        assert b == 2
        assert c == 3

    my_hook = Hook([f, g], kwargs={"c": 3})
    my_hook(1, 2)

    assert call_order == ["f", "g"]



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import product
from typing import Iterable, Optional, Union

import numpy as np
import pytest
import torch

from evotorch import Problem, testing, tools
from evotorch.tools import DType

NON_EMPTY_SHAPES = [(40,), (50, 24), (30, 70, 50)]
SHAPES = [tuple()] + NON_EMPTY_SHAPES
FLOAT_DTYPES = ["float16", "float32", "float64"]
INT_DTYPES = ["int32", "int64"]
DTYPES = FLOAT_DTYPES + INT_DTYPES


FLOAT_TOLERANCE = 1e-4


def _all_must_be_close(x: torch.Tensor, y: torch.Tensor):
    is_float = "float" in str(x.dtype)
    if x.dtype != y.dtype:
        raise testing.TestingError(f"`x` and `y` have different dtypes: {x.dtype}, {y.dtype}")
    if is_float:
        testing.assert_allclose(x, y, atol=FLOAT_TOLERANCE)
    else:
        assert bool(torch.all(x == y))


def _some_must_be_different(x: torch.Tensor, y: torch.Tensor):
    is_float = "float" in str(x.dtype)
    if x.dtype != y.dtype:
        raise testing.TestingError(f"`x` and `y` have different dtypes: {x.dtype}, {y.dtype}")
    if is_float:
        assert torch.any(
            torch.abs(x - y) > FLOAT_TOLERANCE
        ), f"These tensors were expected to be different, but they are almost the same: {x}; {y}"
    else:
        assert torch.any(x != y), f"These tensors were expected to be different, but they are the same: {x}; {y}"


def _all_must_be_between(x: torch.Tensor, lb: Union[float, int, torch.Tensor], ub: Union[float, int, torch.Tensor]):
    is_float = "float" in str(x.dtype)
    if is_float:
        testing.assert_almost_between(x, lb, ub, atol=FLOAT_TOLERANCE)
    else:
        assert bool(torch.all((x >= lb) & (x <= ub)))


def _each_must_be_equal(x: torch.Tensor, value: Union[int, float]):
    is_float = "float" in str(x.dtype)
    if is_float:
        value = float(value)
        if x.ndim == 0:
            assert abs(x - value) <= FLOAT_TOLERANCE
        else:
            testing.assert_eachclose(x, value, atol=FLOAT_TOLERANCE)
    else:
        value = int(value)
        if x.ndim == 0:
            assert x == value
        else:
            assert torch.all(x == value)


def _prepare_dtype(dtype: Union[torch.dtype, str]) -> tuple:
    dtype = tools.to_torch_dtype(dtype)
    is_float = "float" in str(dtype)
    return dtype, is_float


@pytest.mark.parametrize("shape,dtype", product(SHAPES, DTYPES))
def test_basic_makers(shape: tuple, dtype: DType):
    dtype, is_float = _prepare_dtype(dtype)

    empty = tools.make_empty(shape, dtype=dtype)
    assert empty.shape == shape
    assert empty.dtype == dtype

    ones = tools.make_ones(shape, dtype=dtype)
    assert ones.shape == shape
    assert ones.dtype == dtype
    _each_must_be_equal(ones, 1)

    zeros = tools.make_zeros(shape, dtype=dtype)
    assert zeros.shape == shape
    assert zeros.dtype == dtype
    _each_must_be_equal(zeros, 0)

    if is_float:
        nans = tools.make_nan(shape, dtype=dtype)
        assert nans.shape == shape
        assert nans.dtype == dtype
        assert torch.all(torch.isnan(nans))


@pytest.mark.parametrize("shape,dtype", product(NON_EMPTY_SHAPES, DTYPES))
def test_targeted_makers(shape: tuple, dtype: DType):
    dtype, is_float = _prepare_dtype(dtype)

    ones = torch.empty(shape, dtype=dtype)
    tools.make_ones(out=ones)
    _each_must_be_equal(ones, 1)

    zeros = torch.empty(shape, dtype=dtype)
    tools.make_zeros(out=zeros)
    _each_must_be_equal(zeros, 0)

    if is_float:
        nans = torch.empty(shape, dtype=dtype)
        tools.make_nan(out=nans)
        assert torch.all(torch.isnan(nans))


@pytest.mark.parametrize("size,dtype,using_out", product([3, 5, 10], DTYPES, [False, True]))
def test_identity_maker(size: int, dtype: DType, using_out: bool):
    dtype, _ = _prepare_dtype(dtype)

    if using_out:
        I_matrix = torch.empty((size, size), dtype=dtype)
        tools.make_I(out=I_matrix)
    else:
        I_matrix = tools.make_I(size, dtype=dtype)

    desired = torch.eye(size, dtype=dtype)
    _all_must_be_close(I_matrix, desired)


@pytest.mark.parametrize("shape,dtype,using_out", product(SHAPES, FLOAT_DTYPES, [False, True]))
def test_random_maker(shape: tuple, dtype: DType, using_out: bool):
    if using_out and (shape == tuple()):
        return

    dtype, is_float = _prepare_dtype(dtype)

    # The following are the random-sampling functions and some arbitrary configurations for them.
    funcs = [
        (tools.make_uniform, {"lb": -10, "ub": 10}),
        (tools.make_gaussian, {"center": 0, "stdev": 10}),
        (tools.make_randint, {"n": 5}),
    ]

    for func, kwargs in funcs:  # for each random-making function and its keyword arguments

        # The following is a temporary function which samples new tensors using `func`
        def sample(g: Optional[torch.Generator] = None) -> torch.Tensor:
            # Prepare the keyword arguments we will use on the function
            all_kwargs = {}
            all_kwargs.update(kwargs)

            if g is not None:
                # If a generator is provided, then we add the generator to our keyword arguments
                all_kwargs["generator"] = g

            if using_out:
                # We test the `out` keyword argument of the function
                result = torch.empty(shape, dtype=dtype)
                func(out=result, **all_kwargs)
            else:
                # We test the function's capabilities without using `out`.
                result = func(shape, dtype=dtype, **all_kwargs)
                assert result.dtype == dtype  # Make sure that dtype is correct
                assert result.shape == shape  # Make sure that shape is correct

            if ("lb" in all_kwargs) and ("ub" in all_kwargs):
                # If the "lb" and "ub" keyword arguments exist in all_kwargs, then the function we just tested was
                # `make_uniform`. Knowing this, we perform an additional test and ensure that the sample does not
                # violate the boundaries.
                _all_must_be_between(result, all_kwargs["lb"], all_kwargs["ub"])

            return result

        if (len(shape) == 0) and ((not is_float) or ("n" in kwargs)):
            # When we have an empty shape (i.e. when we are sampling a scalar) and when dtype is not float
            # (or when we are using make_randint, which we understand from the existence of "n" in kwargs),
            # then the separate samples can still end up with same values by coincidence, due to limited amount of
            # choices.
            # Therefore, the only test we do here is to just take a single sample and let the inner tests work.
            sample()
        else:
            # Make a random generator
            g = torch.Generator()

            # Set a manual seed, and take two samples
            g.manual_seed(1)
            data1 = sample(g)
            data2 = sample(g)

            # Use the same seed again, and take a sample
            g.manual_seed(1)
            data3 = sample(g)

            # data1 and data2 must be different
            _some_must_be_different(data2, data1)

            # data3 and data1 were sampled from equivalent states of a random generator
            # so, they must be equal
            _all_must_be_close(data3, data1)


def _make_problem(*, dtype: DType, eval_dtype: DType, solution_length: int):
    return Problem(
        "min",
        torch.linalg.norm,
        dtype=dtype,
        eval_dtype=eval_dtype,
        initial_bounds=(-10, 10),
        solution_length=solution_length,
    )


@pytest.mark.parametrize("shape,dtype,eval_dtype", product(SHAPES, DTYPES, FLOAT_DTYPES))
def test_tensor_makers_of_problem(shape: Optional[tuple], dtype: DType, eval_dtype: DType):
    dtype, is_float = _prepare_dtype(dtype)
    eval_dtype, _ = _prepare_dtype(eval_dtype)

    # some arbitrary configurations:
    numsln = 20
    slnlen = 7

    # Make a new Problem instance
    problem = _make_problem(dtype=dtype, eval_dtype=eval_dtype, solution_length=slnlen)

    # some more arbitrary configurations:
    center = 5.0
    stdev = 8.0
    lb = 5
    ub = 10

    # In this test, we consider the following methods of the problem instance.
    # Each tuple is in the form (method, filling_value, further_kwargs).
    # For example, problem.make_ones has its filling_value as 1, which means that it is a method which fills its
    # target with 1 values.
    funcs = [
        (problem.make_empty, None, {}),
        (problem.make_ones, 1, {}),
        (problem.make_zeros, 0, {}),
        (problem.make_nan, float("nan"), {}),
        (problem.make_gaussian, None, {"center": center, "stdev": stdev}),
        (problem.make_uniform, None, {"lb": lb, "ub": ub}),
    ]

    for func, filling_value, further_kwargs in funcs:  # For each function and its filling value
        for use_eval_dtype in (False, True):  # Do without and with use_eval_dtype

            if ("center" in further_kwargs) and ("stdev" in further_kwargs) and (not use_eval_dtype) and (not is_float):
                # If "center" and "stdev" are in further_kwargs, then our function is make_gaussian(...).
                # If our current dtype is an integer type (i.e. not is_float), then make_gaussian(...) cannot work.
                # So, we skip this step of the loop.
                continue

            # Determine the desired dtype of the new tensor
            target_dtype = problem.eval_dtype if use_eval_dtype else problem.dtype

            # Are we trying to use make_gaussian(...) on an integer type (which would fail)?
            gaussian_on_int = (
                ("center" in further_kwargs) and ("stdev" in further_kwargs) and ("int" in str(target_dtype))
            )

            if not gaussian_on_int:
                if (filling_value is not None) and np.isnan(filling_value) and ("int" in str(target_dtype)):
                    # If the new tensor's dtype is expected as an integer, and the function's filling value is NaN, then
                    # we skip this step of the loop.
                    continue

                if shape is None:
                    # If the shape argument is None, then we create a new tensor with the help of the `num_solutions`
                    # keyword argument of the tensor creation method.
                    new_tensor = func(num_solutions=numsln, use_eval_dtype=use_eval_dtype, **further_kwargs)
                    assert new_tensor.shape == (numsln,)  # the new tensor's shape must be correct
                else:
                    # If a shape is given, then we create a new tensor of the desired shape
                    new_tensor = func(shape, use_eval_dtype=use_eval_dtype, **further_kwargs)
                    assert new_tensor.shape == shape  # the new tensor's shape must be correct
                assert new_tensor.dtype == target_dtype  # the new tensor's dtype must be correct

                if filling_value is not None:
                    # If there is a filling value for this function (e.g. like 1 for make_ones, or 0 for make_zeros)
                    # then we need to ensure that the function did the filling correctly.

                    if np.isnan(filling_value):
                        # If the filling value is NaN, then we ensure that all values within the new tensor are NaN
                        assert torch.all(torch.isnan(new_tensor))
                    else:
                        # If the filling value is a real number, then we check if the tensor is filled correctly
                        _each_must_be_equal(new_tensor, filling_value)

                if ("lb" in further_kwargs) and ("ub" in further_kwargs):
                    # If "lb" and "ub" are in the args, then the function we are working with is make_uniform.
                    # With this knowledge, we now ensure that the new tensor has its values within the desired boundaries.
                    _all_must_be_between(new_tensor, lb, ub)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains operators for problems whose solutions contain
variable-length sequences (list-like objects).
"""

from ..core import SolutionBatch
from ..tools.objectarray import ObjectArray
from .base import CrossOver


class CutAndSplice(CrossOver):
    """Cut & Splice operator for variable-length solutions.

    This class serves as a cross-over operator to be used on problems
    with their `dtype`s set as `object`, and with their solutions
    initialized to contain variable-length sequences (list-like objects).

    Reference:

        David E. Goldberg, Bradley Korb, Kalyanmoy Deb (1989).
        Messy Genetic Algorithms: Motivation, Analysis, and First Results.
        Complex Systems 3, 493-530.
    """

    def _cut_and_splice(
        self,
        parents1: ObjectArray,
        parents2: ObjectArray,
        children1: SolutionBatch,
        children2: SolutionBatch,
        row_index: int,
    ):
        parvals1 = parents1[row_index]
        parvals2 = parents2[row_index]

        length1 = len(parvals1)
        length2 = len(parvals2)

        cutpoint1 = int(self.problem.make_randint(tuple(), n=length1))
        cutpoint2 = int(self.problem.make_randint(tuple(), n=length2))

        childvals1 = parvals1[:cutpoint1]
        childvals1.extend(parvals2[cutpoint2:])

        childvals2 = parvals2[:cutpoint2]
        childvals2.extend(parvals1[cutpoint1:])

        children1.access_values(keep_evals=True)[row_index] = childvals1
        children2.access_values(keep_evals=True)[row_index] = childvals2

    def _do_cross_over(self, parents1: ObjectArray, parents2: ObjectArray) -> SolutionBatch:
        n = len(parents1)

        children1 = SolutionBatch(self.problem, popsize=n, empty=True)
        children2 = SolutionBatch(self.problem, popsize=n, empty=True)

        for i in range(n):
            self._cut_and_splice(parents1, parents2, children1, children2, i)

        return children1.concat(children2)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa: C901

"""
This namespace contains the CMAES class, which is a wrapper
for the CMA-ES implementation of the `cma` package.
"""

import math
from copy import copy, deepcopy
from typing import Any, Callable, Iterable, List, Optional, Union

import numpy as np
import torch

from ..core import Problem, SolutionBatch
from ..tools.misc import Device, DType, RealOrVector, Vector, is_sequence, numpy_copy
from .searchalgorithm import SearchAlgorithm, SinglePopulationAlgorithmMixin

try:
    import cma
except ImportError:
    cma = None


class CMAES(SearchAlgorithm, SinglePopulationAlgorithmMixin):
    """
    CMAES: Covariance Matrix Adaptation Evolution Strategy.

    This is an interface class between the CMAES implementation
    within the `cma` package developed within the GitHub repository
    CMA-ES/pycma.

    References:

        Nikolaus Hansen, Youhei Akimoto, and Petr Baudis.
        CMA-ES/pycma on Github. Zenodo, DOI:10.5281/zenodo.2559634,
        February 2019.
        <https://github.com/CMA-ES/pycma>

        Nikolaus Hansen, Andreas Ostermeier (2001).
        Completely Derandomized Self-Adaptation in Evolution Strategies.

    """

    def __init__(
        self,
        problem: Problem,
        *,
        stdev_init: RealOrVector,  # sigma0
        popsize: Optional[int] = None,  # popsize
        center_init: Optional[Vector] = None,  # x0
        center_learning_rate: Optional[float] = None,  # CMA_cmean
        cov_learning_rate: Optional[float] = None,  # CMA_on
        rankmu_learning_rate: Optional[float] = None,  # CMA_rankmu
        rankone_learning_rate: Optional[float] = None,  # CMA_rankone
        stdev_min: Optional[Union[float, np.ndarray]] = None,  # minstd
        stdev_max: Optional[Union[float, np.ndarray]] = None,  # maxstd
        separable: bool = False,  # CMA_diagonal
        obj_index: Optional[int] = None,
        cma_options: dict = {},
    ):
        """
        `__init__(...)`: Initialize the CMAES solver.

        Args:
            problem: The problem object which is being worked on.
            stdev_init: Initial standard deviation as a scalar or
                as a 1-dimensional array.
            popsize: Population size. Can be specified as an int,
                or can be left as None to let the CMAES solver
                decide the population size according to the length
                of a solution.
            center_init: Initial center point of the search distribution.
                Can be given as a SolutionVector or as a 1-D array.
                If left as None, an initial center point is generated
                with the help of the problem object's `generate_values(...)`
                method.
            center_learning_rate: Learning rate for updating the mean
                of the search distribution. Leaving this as None
                means that the CMAES solver is to use its own default,
                which is documented as 1.0.
            cov_learning_rate: Learning rate for updating the covariance
                matrix of the search distribution. This hyperparameter
                acts as a common multiplier for rank_one update and rank_mu
                update of the covariance matrix. Leaving this as None
                means that the CMAES solver is to use its own default,
                which is documented as 1.0.
            rankmu_learning_rate: Learning rate for the rank_mu update
                of the covariance matrix of the search distribution.
                Leaving this as None means that the CMAES solver is to use
                its own default, which is documented as 1.0.
            rankone_learning_rate: Learning rate for the rank_one update
                of the covariance matrix of the search distribution.
                Leaving this as None means that the CMAES solver is to use
                its own default, which is documented as 1.0.
            stdev_min: Minimum allowed standard deviation of the search
                distribution. Leaving this as None means that no such
                boundary is to be used.
                Can be given as None, as a scalar, or as a 1-dimensional
                array.
            stdev_max: Maximum allowed standard deviation of the search
                distribution. Leaving this as None means that no such
                boundary is to be used.
                Can be given as None, as a scalar, or as a 1-dimensional
                array.
            separable: Provide this as True if you would like the problem
                to be treated as a separable one. Treating a problem
                as separable means to adapt only the diagonal parts
                of the covariance matrix and to keep the non-diagonal
                parts 0. High dimensional problems result in large
                covariance matrices on which operating is computationally
                expensive. Therefore, for such high dimensional problems,
                setting `separable` as True might be useful.
                If, instead, you would like to configure on which
                iterations the diagonal parts of the covariance matrix
                are to be adapted, then it is recommended to leave
                `separable` as False and set a new value for the key
                "CMA_diagonal" via `cma_options` (see the official
                documentation of pycma for details regarding the
                "CMA_diagonal" setting).
            obj_index: Objective index according to which evaluation
                of the solution will be done.
            cma_options: Any other configuration for the CMAES solver
                can be passed via the cma_options dictionary.
        """

        # Make sure that the cma module is installed
        if cma is None:
            raise ImportError(f"The class {type(self).__name__} is only available if the package `cma` is installed.")

        # Initialize the base class
        SearchAlgorithm.__init__(self, problem, center=self._get_center)

        # Initialize the population.
        self._population: SolutionBatch = self._problem.generate_batch(popsize, empty=True)

        # Ensure that the problem is numeric
        problem.ensure_numeric()

        # Store the objective index
        self._obj_index = problem.normalize_obj_index(obj_index)

        # If `center_init` is not given, generate an initial solution
        # with the help of the problem object.
        # Otherwise, use the given initial solution as the starting
        # point in the search space.
        if center_init is None:
            x0 = self._problem.generate_values(1).to("cpu").view(-1).numpy().astype(dtype=float)
        else:
            x0 = numpy_copy(center_init, dtype=float)

        # Store the initial standard deviations
        sigma0 = numpy_copy(stdev_init, dtype=float)

        # Generate an options dictionary to pass to the cma solver.
        inopts = {}
        for k, v in cma_options.items():
            if isinstance(v, torch.Tensor):
                v = numpy_copy(v, dtype=float)
            inopts[k] = v

        # Remove the number of iterations boundary
        if "maxiter" not in inopts:
            inopts["maxiter"] = np.inf

        # Below is a temporary helper function for safely storing the configuration items.
        # This inner function updates the `inopts` variable.
        def store_opt(key: str, long_name: str, value: Any, converter: Callable):
            # Here, `key` represents the configuration key used by pycma
            # `long_name` represents the configuration's long name used by this class
            # `value` is the configuration value associated with `key`.

            # Declare that this inner function accesses the `inopts` variable.
            nonlocal inopts

            if value is None:
                # If the provided `value` is None, then there is no configuration to store.
                # So, we just leave this inner function.
                return

            if key in inopts:
                # If the given `key` already exists within `inopts`, this means that the configuration was specified
                # twice: via the keyword argument `cma_options` AND via a keyword argument.
                # We raise an error and inform the user about this redundancy.
                raise ValueError(
                    f"The configuration {repr(key)} was redundantly provided"
                    f" both via the initialization argument {long_name}"
                    f" and via the cma_options dictionary."
                    f" {long_name}={repr(value)};"
                    f" cma_options[{repr(key)}]={repr(inopts[key])}."
                )

            inopts[key] = converter(value)

        # Temporary helper function which makes sure that `x` is a numpy array or a float.
        def array_or_float(x):
            if is_sequence(x):
                return numpy_copy(x)
            else:
                return float(x)

        # Store the cma configuration received through the initialization arguments (and raise error if there is
        # redundancy with the cma_options dictionary).
        store_opt("popsize", "popsize", popsize, int)
        store_opt("CMA_cmean", "center_learning_rate", center_learning_rate, float)
        store_opt("CMA_on", "cov_learning_rate", cov_learning_rate, float)
        store_opt("CMA_rankmu", "rankmu_learning_rate", rankmu_learning_rate, float)
        store_opt("CMA_rankone", "rankone_learning_rate", rankone_learning_rate, float)
        store_opt("minstd", "stdev_min", stdev_min, array_or_float)
        store_opt("maxstd", "stdev_max", stdev_max, array_or_float)
        if separable:
            store_opt("CMA_diagonal", "separable", separable, bool)

        # If the problem defines lower and upper bounds, pass these into the options dict.
        def process_bounds(bounds: RealOrVector) -> np.ndarray:
            if bounds is None:
                return None
            else:
                if is_sequence(bounds):
                    bounds = numpy_copy(bounds)
                else:
                    bounds = np.array(float(bounds)).repeat(self._problem.solution_length)
                return bounds

        lb = process_bounds(self._problem.lower_bounds)
        ub = process_bounds(self._problem.upper_bounds)

        register_bounds = False
        if lb is not None and ub is None:
            ub = np.array(np.inf).repeat(self._problem.solution_length)
            register_bounds = True
        elif lb is None and ub is not None:
            lb = np.array(-(np.inf)).repeat(self._problem.solution_length)
            register_bounds = True
        elif lb is not None and ub is not None:
            register_bounds = True

        if register_bounds:
            inopts["bounds"] = [lb, ub]

        # Generate a random seed using the problem object for the sake of reproducibility.
        if "seed" not in inopts:
            inopts["seed"] = int(self._problem.make_randint(tuple(), n=(2**32) - 100) + 100)

        # Instantiate the CMAEvolutionStrategy with the prepared configuration items.
        self._es = cma.CMAEvolutionStrategy(x0, sigma0, inopts)

        # Use the SinglePopulationAlgorithmMixin to enable additional status reports regarding the population.
        SinglePopulationAlgorithmMixin.__init__(self)

    @property
    def population(self) -> SolutionBatch:
        """Population generated by the CMA-ES algorithm"""
        return self._population

    def _step(self):
        """Perform a step of the CMA-ES solver"""
        asked = self._es.ask()
        self._population.access_values()[:] = torch.as_tensor(
            np.asarray(asked), dtype=self._problem.dtype, device=self._population.device
        )
        self._problem.evaluate(self._population)
        scores = numpy_copy(self._population.utility(self._obj_index), dtype=float)
        self._es.tell(asked, -1.0 * scores)

    def _get_center(self) -> torch.Tensor:
        return torch.as_tensor(self._es.result[5], dtype=self._population.dtype, device=self._population.device)

    @property
    def obj_index(self) -> int:
        """Index of the objective being focused on"""
        return self._obj_index



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimizers (like Adam or ClipUp) to be used with distribution-based
search algorithms.
"""

import math
from typing import Callable, Optional, Type

import torch
import torch.nn.utils as nnu

from .tools.misc import Device, DType, RealOrVector, ensure_tensor_length_and_dtype, to_torch_dtype


class TorchOptimizer:
    """
    Base class for using a PyTorch optimizer
    """

    def __init__(
        self,
        torch_optimizer: Type,
        *,
        config: dict,
        solution_length: int,
        dtype: DType,
        device: Device = "cpu",
    ):
        """
        `__init__(...)`: Initialize the TorchOptimizer.

        Args:
            torch_optimizer: The class which represents a PyTorch optimizer.
            config: The configuration dictionary to be passed to the optimizer
                as keyword arguments.
            solution_length: Length of a solution of the problem on which the
                optimizer will work.
            dtype: The dtype of the problem.
            device: The device on which the solutions are kept.
        """
        self._data = torch.empty(int(solution_length), dtype=to_torch_dtype(dtype), device=device)
        self._optim = torch_optimizer([self._data], **config)

    @torch.no_grad()
    def ascent(self, globalg: RealOrVector, *, cloned_result: bool = True) -> torch.Tensor:
        """
        Compute the ascent, i.e. the step to follow.

        Args:
            globalg: The estimated gradient.
            cloned_result: If `cloned_result` is True, then the result is a
                copy, guaranteed not to be the view of any other tensor
                internal to the TorchOptimizer class.
                If `cloned_result` is False, then the result is not a copy.
                Use `cloned_result=False` only when you are sure that your
                algorithm will never do direct modification on the ascent
                vector it receives.
        Returns:
            The ascent vector, representing the step to follow.
        """

        globalg = ensure_tensor_length_and_dtype(
            globalg,
            len(self._data),
            dtype=self._data.dtype,
            device=self._data.device,
            about=f"{type(self).__name__}.ascent",
        )

        self._data.zero_()
        self._data.grad = globalg
        self._optim.step()
        result = -1.0 * self._data

        return result


class Adam(TorchOptimizer):
    """
    The Adam optimizer.

    Reference:

        Kingma, D. P. and J. Ba (2015).
        Adam: A method for stochastic optimization.
        In Proceedings of 3rd International Conference on Learning Representations.
    """

    def __init__(
        self,
        *,
        solution_length: int,
        dtype: DType,
        device: Device = "cpu",
        stepsize: Optional[float] = None,
        beta1: Optional[float] = None,
        beta2: Optional[float] = None,
        epsilon: Optional[float] = None,
        amsgrad: Optional[bool] = None,
    ):
        """
        `__init__(...)`: Initialize the Adam optimizer.

        Args:
            solution_length: Length of a solution of the problem which is
                being worked on.
            dtype: The dtype of the problem which is being worked on.
            device: The device on which the solutions are kept.
            stepsize: The step size (i.e. the learning rate) employed
                by the optimizer.
            beta1: The beta1 hyperparameter. None means the default.
            beta2: The beta2 hyperparameter. None means the default.
            epsilon: The epsilon hyperparameters. None means the default.
            amsgrad: Whether or not to use the amsgrad behavior.
                None means the default behavior.
                See `torch.optim.Adam` for details.
        """

        config = {}

        if stepsize is not None:
            config["lr"] = float(stepsize)

        if beta1 is None and beta2 is None:
            pass  # nothing to do
        elif beta1 is not None and beta2 is not None:
            config["betas"] = (float(beta1), float(beta2))
        else:
            raise ValueError(
                "The arguments beta1 and beta2 were expected"
                " as both None, or as both real numbers."
                " However, one of them was encountered as None and"
                " the other was encountered as something other than None."
            )

        if epsilon is not None:
            config["eps"] = float(epsilon)

        if amsgrad is not None:
            config["amsgrad"] = bool(amsgrad)

        super().__init__(torch.optim.Adam, solution_length=solution_length, dtype=dtype, device=device, config=config)


class SGD(TorchOptimizer):
    """
    The SGD optimizer.

    Reference regarding the momentum behavior:

        Polyak, B. T. (1964).
        Some methods of speeding up the convergence of iteration methods.
        USSR Computational Mathematics and Mathematical Physics, 4(5):1–17.

    Reference regarding the Nesterov behavior:

        Yurii Nesterov (1983).
        A method for unconstrained convex minimization problem with the rate ofconvergence o(1/k2).
        Doklady ANSSSR (translated as Soviet.Math.Docl.), 269:543–547.
    """

    def __init__(
        self,
        *,
        solution_length: int,
        dtype: DType,
        stepsize: float,
        device: Device = "cpu",
        momentum: Optional[float] = None,
        dampening: Optional[bool] = None,
        nesterov: Optional[bool] = None,
    ):
        """
        `__init__(...)`: Initialize the SGD optimizer.

        Args:
            solution_length: Length of a solution of the problem which is
                being worked on.
            dtype: The dtype of the problem which is being worked on.
            stepsize: The step size (i.e. the learning rate) employed
                by the optimizer.
            device: The device on which the solutions are kept.
            momentum: The momentum coefficient. None means the default.
            dampening: Whether or not to activate the dampening behavior.
                None means the default.
                See `torch.optim.SGD` for details.
            nesterov: Whether or not to activate the nesterov behavior.
                None means the default.
                See `torch.optim.SGD` for details.
        """

        config = {}

        config["lr"] = float(stepsize)

        if momentum is not None:
            config["momentum"] = float(momentum)

        if dampening is not None:
            config["dampening"] = float(dampening)

        if nesterov is not None:
            config["nesterov"] = bool(nesterov)

        super().__init__(torch.optim.SGD, solution_length=solution_length, dtype=dtype, device=device, config=config)


class ClipUp:
    """
    The ClipUp optimizer.

    Although this optimizer has the very same interface with SGD and Adam,
    it is not a PyTorch optimizer. Therefore, it does not inherit from
    TorchOptimizer.

    Reference:

        Toklu, N. E., Liskowski, P., & Srivastava, R. K. (2020, September).
        ClipUp: A Simple and Powerful Optimizer for Distribution-Based Policy Evolution.
        In International Conference on Parallel Problem Solving from Nature (pp. 515-527).
        Springer, Cham.
    """

    def __init__(
        self,
        *,
        solution_length: int,
        dtype: DType,
        stepsize: float,
        momentum: float = 0.9,
        max_speed: Optional[float] = None,
        device: Device = "cpu",
    ):
        """
        `__init__(...)`: Initialize the ClipUp optimizer.

        Args:
            solution_length: Length of a solution of the problem which is
                being worked on.
            dtype: The dtype of the problem which is being worked on.
            stepsize: The step size (i.e. the learning rate) employed
                by the optimizer.
            momentum: The momentum coefficient. None means the default.
            max_speed: The maximum speed. If given as None, the
                `max_speed` will be taken as two times the stepsize.
            device: The device on which the solutions are kept.
        """

        stepsize = float(stepsize)
        momentum = float(momentum)
        if max_speed is None:
            max_speed = stepsize * 2.0
        else:
            max_speed = float(max_speed)
        solution_length = int(solution_length)

        if stepsize < 0.0:
            raise ValueError(f"Invalid stepsize: {stepsize}")
        if momentum < 0.0 or momentum > 1.0:
            raise ValueError(f"Invalid momentum: {momentum}")
        if max_speed < 0.0:
            raise ValueError(f"Invalid max_speed: {max_speed}")

        self._stepsize = stepsize
        self._momentum = momentum
        self._max_speed = max_speed

        self._velocity: Optional[torch.Tensor] = torch.zeros(
            solution_length, dtype=to_torch_dtype(dtype), device=device
        )

        self._dtype = to_torch_dtype(dtype)
        self._device = device

    @staticmethod
    def _clip(x: torch.Tensor, limit: float) -> torch.Tensor:
        with torch.no_grad():
            normx = torch.norm(x)
            if normx > limit:
                ratio = limit / normx
                return x * ratio
            else:
                return x

    @torch.no_grad()
    def ascent(self, globalg: RealOrVector, *, cloned_result: bool = True) -> torch.Tensor:
        """
        Compute the ascent, i.e. the step to follow.

        Args:
            globalg: The estimated gradient.
            cloned_result: If `cloned_result` is True, then the result is a
                copy, guaranteed not to be the view of any other tensor
                internal to the TorchOptimizer class.
                If `cloned_result` is False, then the result is not a copy.
                Use `cloned_result=False` only when you are sure that your
                algorithm will never do direct modification on the ascent
                vector it receives.
                Important: if you set `cloned_result=False`, and do in-place
                modifications on the returned result of `ascent(...)`, then
                the internal velocity of ClipUp will be corrupted!
        Returns:
            The ascent vector, representing the step to follow.
        """

        globalg = ensure_tensor_length_and_dtype(
            globalg,
            len(self._velocity),
            dtype=self._dtype,
            device=self._device,
            about=f"{type(self).__name__}.ascent",
        )

        grad = (globalg / torch.norm(globalg)) * self._stepsize

        self._velocity = self._clip((self._momentum * self._velocity) + grad, self._max_speed)

        result = self._velocity

        if cloned_result:
            result = result.clone()

        return result


def get_optimizer_class(s: str, optimizer_config: Optional[dict] = None) -> Callable:
    """
    Get the optimizer class from the given string.

    Args:
        s: A string, referring to the optimizer class.
            "clipsgd", "clipsga", "clipup" refers to ClipUp.
            "adam" refers to Adam.
            "sgd" or "sga" refers to SGD.
        optimizer_config: A dictionary containing the configurations to be
            passed to the optimizer. If this argument is not None,
            then, instead of the class being referred to, a dynamically
            generated factory function will be returned, which will pass
            these configurations to the actual class upon being called.
    Returns:
        The class, or a factory function instantiating that class.
    """
    if s in ("clipsgd", "clipsga", "clipup"):
        cls = ClipUp
    elif s == "adam":
        cls = Adam
    elif s in ("sgd", "sga"):
        cls = SGD
    else:
        raise ValueError(f"Unknown optimizer: {repr(s)}")

    if optimizer_config is None:
        return cls
    else:

        def f(*args, **kwargs):
            nonlocal cls, optimizer_config
            conf = {}
            conf.update(optimizer_config)
            conf.update(kwargs)
            return cls(*args, **conf)

        return f



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy, deepcopy
from itertools import product
from typing import Callable, Iterable, Optional, Union

import pytest
import torch

import evotorch as et
import evotorch.tools as ett
from evotorch import testing
from evotorch.tools.objectarray import ObjectArray


class CloningMethods:
    def clone_via_method(x):
        return x.clone()

    def clone_via_copy(x):
        return copy(x)

    def clone_via_deepcopy(x):
        return deepcopy(x)

    def clone_via_clone_func(x):
        return ett.clone(x)


class DummyProblems:
    @staticmethod
    def _sum(x: torch.Tensor) -> torch.Tensor:
        if x.ndim == 1:
            return torch.sum(x)
        elif x.ndim == 2:
            return torch.sum(x, dim=-1)
        else:
            raise ValueError(f"Unexpected shape: {x.shape}")

    @staticmethod
    def _make_simple_problem(
        *,
        vectorized: bool,
        objective_sense: Union[str, Iterable[str]] = "min",
        solution_length: int = 5,
        dtype: ett.DType = "float32",
        num_actors: Union[int, str] = 0,
    ):
        return et.Problem(
            objective_sense,
            DummyProblems._sum,
            solution_length=solution_length,
            dtype=dtype,
            num_actors=num_actors,
            initial_bounds=(-10.0, 10.0),
            vectorized=vectorized,
        )

    @staticmethod
    def make_simple_non_batched_problem(
        *,
        objective_sense: Union[str, Iterable[str]] = "min",
        solution_length: int = 5,
        dtype: ett.DType = "float32",
        num_actors: Union[int, str] = 0,
    ):
        return DummyProblems._make_simple_problem(
            objective_sense=objective_sense,
            solution_length=solution_length,
            dtype=dtype,
            num_actors=num_actors,
            vectorized=False,
        )

    @staticmethod
    def make_simple_batched_problem(
        *,
        objective_sense: Union[str, Iterable[str]] = "min",
        solution_length: int = 5,
        dtype: ett.DType = "float32",
        num_actors: Union[int, str] = 0,
    ):
        return DummyProblems._make_simple_problem(
            objective_sense=objective_sense,
            solution_length=solution_length,
            dtype=dtype,
            num_actors=num_actors,
            vectorized=True,
        )

    class Batched(et.Problem):
        def __init__(
            self,
            *,
            objective_sense: Union[str, Iterable[str]] = "min",
            solution_length: int = 5,
            dtype: ett.DType = "float32",
            num_actors: Union[int, str] = 0,
            eval_data_length: Optional[int] = None,
        ):
            super().__init__(
                objective_sense=objective_sense,
                solution_length=5,
                num_actors=num_actors,
                dtype=dtype,
                initial_bounds=(-10.0, 10.0),
                eval_data_length=eval_data_length,
            )

        def _evaluate_batch(self, batch: et.SolutionBatch):
            batch.forget_evals()
            batch.access_evals(0)[:] = torch.sum(batch.access_values(), dim=-1)

    class NonBatched(et.Problem):
        def __init__(
            self,
            *,
            objective_sense: Union[str, Iterable[str]] = "min",
            solution_length: int = 5,
            dtype: ett.DType = "float32",
            num_actors: Union[int, str] = 0,
        ):
            super().__init__(
                objective_sense=objective_sense,
                solution_length=5,
                num_actors=num_actors,
                dtype=dtype,
                initial_bounds=(-10.0, 10.0),
            )

        def _evaluate(self, x: et.Solution):
            x.set_evaluation(torch.sum(x.access_values()))

    class MultiObj(et.Problem):
        def __init__(self, *, eval_data_length: Optional[int] = None):
            super().__init__(
                objective_sense=["min", "max"],
                solution_length=2,
                dtype="float32",
                initial_bounds=(-10.0, 10.0),
                eval_data_length=eval_data_length,
            )

        def _evaluate_batch(self, batch: et.SolutionBatch):
            values = batch.access_values()
            x = values[:, 0]
            y = values[:, 1]
            batch.forget_evals()
            batch.access_evals(0)[:] = x
            batch.access_evals(1)[:] = y

    class OptimizeStrings(et.Problem):
        def __init__(self):
            super().__init__(objective_sense="min", dtype=object)

        def _evaluate(self, x: et.Solution):
            s = x.access_values()
            assert isinstance(s, str)
            x.set_evaluation(float(s))

        def _fill(self, x: ObjectArray):
            for i in range(len(x)):
                x[i] = str(int(self.make_randint(tuple(), n=100)))


@pytest.mark.parametrize(
    "prob_class,num_actors",
    [
        (DummyProblems.make_simple_batched_problem, 0),
        (DummyProblems.make_simple_non_batched_problem, 0),
        (DummyProblems.Batched, 0),
        (DummyProblems.NonBatched, 0),
        (DummyProblems.make_simple_batched_problem, 2),
        (DummyProblems.make_simple_non_batched_problem, 2),
        (DummyProblems.Batched, 2),
        (DummyProblems.NonBatched, 2),
    ],
)
def test_batch_and_solutions(prob_class, num_actors):
    # Instantiate the problem
    prob = prob_class(num_actors=num_actors)

    # Declare a population size
    popsize = 10

    # Make a batch of the specified population size
    batch: et.SolutionBatch = prob.generate_batch(popsize)

    # Length and shape of the batch should be consistent
    assert len(batch) == popsize
    assert batch.values_shape[0] == popsize
    assert batch.eval_shape[0] == popsize

    # Evaluate the batch
    prob.evaluate(batch)

    # The example problems in this test return the sum of decision values as fitnesses.
    # We manually compare the values here and assert that the fitnesses are correct.
    evals = batch.access_evals(0)
    for i in range(popsize):
        f = evals[i]
        s = torch.sum(batch.access_values(keep_evals=True)[i, :])
        testing.assert_allclose(s, f, atol=0.00001)

    # Here, we call `access_values()`, which should clear the attached fitnesses.
    batch.access_values()

    # Assert that the fitnesses are erased (they now should be nan).
    evals = batch.access_evals(0)
    assert torch.all(torch.isnan(evals))


def test_batch_slicing():
    prob = DummyProblems.Batched()
    popsize = 10

    # Make a new SolutionBatch
    batch = prob.generate_batch(10)

    # Evaluate the batch
    prob.evaluate(batch)

    # We will slice this batch from its middle row
    # (which we call the cutpoint in this test)
    cutpoint = popsize // 2

    # Take the bottom half of the population
    batchview = batch[cutpoint:]

    # Accessing the values of this batch should affect the original batch too.
    # Because `batchview` is just a view of the original batch.
    batchview.access_values()[:] = 1.0

    # Let's see if the related fitnesses in the original batch are indeed
    # erased.
    org_fitnesses = batch.access_evals(0)[cutpoint:]
    assert torch.all(torch.isnan(org_fitnesses))

    # Let's see if the decision values in the original batch got updated too.
    testing.assert_allclose(
        batch.access_values(keep_evals=True)[cutpoint:], batchview.access_values(keep_evals=True), atol=0.00001
    )

    # Now, let us fill the original batch with ones
    batch.access_values()[:] = 1.0

    # Let us also make sure that the original batch is evaluated again.
    prob.evaluate(batch)

    # Let's make another slice, but this time using an indices list.
    # Like in PyTorch, an indices list is considered an advanced slicing,
    # and it should lead to a copy, not a view.
    batch2 = batch[[1, 4, 2]]

    # Let's make all the decision values in the new batch 0.0
    batch2.access_values()[:] = 0.0

    # Did we cause the original batch's fitnesses to be erased?
    # Hopefully, we did not.
    assert not torch.any(torch.isnan(batch.access_evals(0)))

    # We now evaluate the second batch
    prob.evaluate(batch2)

    # Between the original batch (batch) and the newly made batch (batch2),
    # both the decision values and the fitnesses should be different at this point.
    bvalues = batch.access_values(keep_evals=True)
    b2values = batch2.access_values(keep_evals=True)

    bfitnesses = batch.access_evals(0)
    b2fitnesses = batch2.access_evals(0)

    testing.assert_allclose(bvalues, torch.ones_like(bvalues), atol=0.00001)
    testing.assert_allclose(b2values, torch.zeros_like(b2values), atol=0.00001)

    testing.assert_allclose(bfitnesses, torch.ones_like(bfitnesses) * batch.solution_length, atol=0.00001)
    testing.assert_allclose(b2fitnesses, torch.zeros_like(b2fitnesses), atol=0.00001)


@pytest.mark.parametrize(
    "clone_func",
    [
        CloningMethods.clone_via_clone_func,
        CloningMethods.clone_via_copy,
        CloningMethods.clone_via_deepcopy,
        CloningMethods.clone_via_method,
    ],
)
def test_batch_cloning(clone_func):
    prob = DummyProblems.Batched()

    # Create a new SolutionBatch
    batch = prob.generate_batch(10)

    # Create a new clone of the original SolutionBatch
    batch2 = clone_func(batch)

    # We fill the original batch with ones, and the clone with zeroes.
    batch.access_values()[:] = 1.0
    batch2.access_values()[:] = 0.0

    # While batch2 gets filled with zeroes, the original batch should stay
    # the same (filled with ones). Below, we do the assertion for this.
    bvalues = batch.access_values(keep_evals=True)
    b2values = batch2.access_values(keep_evals=True)
    testing.assert_allclose(bvalues, torch.ones_like(bvalues), atol=0.00001)
    testing.assert_allclose(b2values, torch.zeros_like(b2values), atol=0.00001)


def _storage_address(solution_or_batch: Union[et.SolutionBatch, et.Solution]) -> int:
    return solution_or_batch.access_values(keep_evals=True).storage().data_ptr()


def _share_memory(a: Union[et.SolutionBatch, et.Solution], b: Union[et.SolutionBatch, et.Solution]) -> bool:
    return _storage_address(a) == _storage_address(b)


def test_solutions_referring_to_batch():
    prob = DummyProblems.Batched()

    # Create a new SolutionBatch, consisting of two solutions.
    batch = prob.generate_batch(2)

    # Fill the batch with 10s.
    batch.access_values()[:] = 10.0

    # Evaluate the new batch
    prob.evaluate(batch)

    # Get the solution length
    slnlen = prob.solution_length

    # Get the two solutions from the batch
    sln1 = batch[0]
    sln2 = batch[1]

    # Since we obtained these solutions from a SolutionBatch, these Solution
    # objects are references to the batch (i.e. they share their memories
    # with the batch).
    assert _share_memory(sln1, batch)
    assert _share_memory(sln2, batch)

    # Are the lengths correct?
    assert len(sln1) == slnlen
    assert len(sln2) == slnlen
    assert sln1.shape == (slnlen,)
    assert sln2.shape == (slnlen,)

    # Fill the first solution with zeroes
    sln1.access_values()[:] = 0.0

    # Accessing the values of the first solution should have caused the
    # fitness of the related row in the batch to be erased.
    # Also, the decision values of that first solution in the batch should
    # reflect the changes.
    assert torch.isnan(batch.access_evals(0)[0])
    # assert sln1.evaluation is None
    assert not sln1.is_evaluated
    testing.assert_eachclose(batch.access_values(keep_evals=True)[0, :], 0.0, atol=0.00001)

    # The other solution in the batch should stay the same.
    testing.assert_allclose(batch.access_evals(0)[1], 10 * slnlen, atol=0.00001)
    testing.assert_allclose(sln2.evaluation, 10 * slnlen, atol=0.00001)
    assert sln2.is_evaluated
    testing.assert_eachclose(batch.access_values(keep_evals=True)[1, :], 10.0, atol=0.00001)

    # Fill the second solution with 5s.
    sln2.access_values()[:] = 5.0

    # Now, the fitness of the second solution should be gone as well.
    assert torch.isnan(batch.access_evals(0)[1])
    # assert sln2.evaluation is None
    assert not sln2.is_evaluated
    testing.assert_eachclose(batch.access_values(keep_evals=True)[1, :], 5.0, atol=0.00001)


@pytest.mark.parametrize(
    "clone_func",
    [
        CloningMethods.clone_via_clone_func,
        CloningMethods.clone_via_copy,
        CloningMethods.clone_via_deepcopy,
        CloningMethods.clone_via_method,
    ],
)
def test_solution_cloning(clone_func):
    prob = DummyProblems.Batched()

    # Create a new SolutionBatch, consisting of two solutions.
    batch = prob.generate_batch(2)

    # Fill the batch with 10s.
    batch.access_values()[:] = 10.0

    # Evaluate the new batch
    prob.evaluate(batch)

    # Get a new solution
    sln = batch[0]
    assert _share_memory(batch, sln)

    # Clone the solution
    sln2: et.Solution = clone_func(sln)

    # Since the new solution is a clone, it should NOT be a reference anymore
    assert not _share_memory(sln, sln2)

    sln2.access_values()[:] = 0.0

    testing.assert_eachclose(sln2, 0.0, atol=0.00001)
    testing.assert_eachclose(batch.access_values(keep_evals=True)[:], 10.0, atol=0.00001)


@pytest.mark.parametrize(
    "sense,ranking_method", [("min", None), ("max", None), ("min", "centered"), ("max", "centered")]
)
def test_utilities(sense, ranking_method):
    prob = DummyProblems.Batched(objective_sense=sense)
    batch = prob.generate_batch(2)

    values = batch.access_values()
    numel = values.numel()
    values.view(-1)[:] = torch.arange(numel, dtype=values.dtype)

    prob.evaluate(batch)

    utils = batch.utility()

    if sense == "min":
        assert utils[0] > utils[1]
    else:
        assert utils[1] > utils[0]


def test_multiobj():
    prob = DummyProblems.MultiObj()
    # Objectives: min, max

    batch = et.SolutionBatch(prob, popsize=5, empty=True)
    batch.access_values()[:] = torch.tensor(
        [
            [10, 2],
            [2, 10],
            [1, 9],
            [3, 7],
            [4, 8],
        ],
        dtype=batch.values_dtype,
    )

    prob.evaluate(batch)

    pareto = batch.arg_pareto_sort()

    assert torch.all(pareto.ranks == torch.LongTensor([2, 0, 0, 1, 1]))

    for ranking_method in (None, "centered"):
        utils = batch.utils(ranking_method=ranking_method)

        assert utils.shape == (5, 2)

        first_obj_utils = utils[:, 0]
        first_obj_sorted = first_obj_utils.argsort(descending=True)
        assert torch.all(first_obj_sorted == torch.LongTensor([2, 1, 3, 4, 0]))

        second_obj_utils = utils[:, 1]
        second_obj_sorted = second_obj_utils.argsort(descending=True)
        assert torch.all(second_obj_sorted == torch.LongTensor([1, 2, 4, 3, 0]))


def test_prob_with_dtype_object():
    prob = DummyProblems.OptimizeStrings()

    batch = prob.generate_batch(5)
    prob.evaluate(batch)

    fitnesses = [float(solution.access_values(keep_evals=True)) for solution in batch]

    testing.assert_allclose(batch.access_evals(0), fitnesses, atol=0.00001)


def test_api():
    def f(x: torch.Tensor):
        return torch.sum(x)

    def fmulti(x: torch.Tensor):
        return torch.sum(x), torch.linalg.norm(x)

    with pytest.raises(ValueError):
        et.Problem("min", f)  # Missing solution_length and (initial_)bounds

    with pytest.raises(ValueError):
        et.Problem("min", f, solution_length=5)  # Missing (initial_)bounds

    with pytest.raises(ValueError):
        # This should fail, because et.Problem cannot work with non-numeric dtypes
        # when solution_length and bounds are given
        et.Problem("min", f, solution_length=5, bounds=(-1.0, 1.0), dtype=object)

    prob = et.Problem("min", f, solution_length=5, bounds=(-1.0, 1.0))
    prob.ensure_numeric()
    prob.ensure_single_objective()

    with pytest.raises(ValueError):
        prob.ensure_unbounded()

    assert prob.objective_sense == "min"
    assert prob.senses == ["min"]

    assert prob.initial_lower_bounds is not None
    assert prob.initial_upper_bounds is not None
    testing.assert_allclose(prob.lower_bounds, prob.initial_lower_bounds, atol=0.00001)
    testing.assert_allclose(prob.upper_bounds, prob.initial_upper_bounds, atol=0.00001)

    prob2 = et.Problem("min", f, solution_length=5, initial_bounds=(-1.0, 1.0), bounds=(-5.0, 5.0))
    prob2.ensure_numeric()
    prob2.ensure_single_objective()

    with pytest.raises(ValueError):
        prob2.ensure_unbounded()

    assert prob2.objective_sense == "min"
    assert prob2.senses == ["min"]

    prob3 = et.Problem("max", f, solution_length=5, initial_bounds=(-1.0, 1.0))
    prob3.ensure_unbounded()
    prob3.ensure_numeric()
    prob3.ensure_single_objective()

    assert prob3.objective_sense == "max"
    assert prob3.senses == ["max"]

    prob4 = et.Problem(["min", "min"], fmulti, solution_length=5, initial_bounds=(-1.0, 1.0))
    prob4.ensure_unbounded()
    prob4.ensure_numeric()

    with pytest.raises(ValueError):
        prob4.ensure_single_objective()

    assert prob4.objective_sense == ["min", "min"]
    assert prob4.senses == ["min", "min"]


def test_problem_hooks():
    class FuncWasCalled:
        f1 = False
        f2 = False

    prob = DummyProblems.Batched()

    def f1(b):
        assert isinstance(b, et.SolutionBatch)
        FuncWasCalled.f1 = True

    def f2(b):
        assert isinstance(b, et.SolutionBatch)
        FuncWasCalled.f2 = True
        return {"dummy_status": 1}

    prob.before_eval_hook.append(f1)
    prob.after_eval_hook.append(f2)

    batch = prob.generate_batch(5)
    prob.evaluate(batch)

    assert FuncWasCalled.f1
    assert FuncWasCalled.f2

    assert "dummy_status" in prob.status
    assert prob.status["dummy_status"] == 1


def _must_be_regular_tensor(x: torch.Tensor):
    assert isinstance(x, torch.Tensor) and (not isinstance(x, ett.ReadOnlyTensor))


def _must_be_read_only_tensor(x: torch.Tensor):
    assert isinstance(x, torch.Tensor) and isinstance(x, ett.ReadOnlyTensor)


def test_accessing():
    prob = DummyProblems.NonBatched()
    length = prob.solution_length
    population = prob.generate_batch(10)
    n = len(population)

    _must_be_regular_tensor(population.access_values(keep_evals=True))
    _must_be_regular_tensor(population.access_values())
    _must_be_read_only_tensor(population.values)
    _must_be_read_only_tensor(population.evals)
    testing.assert_shape_matches(population.access_values(keep_evals=True), (n, length))
    testing.assert_shape_matches(population.access_evals(), (n, 1))
    testing.assert_shape_matches(population.values, (n, length))
    testing.assert_shape_matches(population.evals, (n, 1))

    sln = population[0]
    _must_be_regular_tensor(sln.access_values(keep_evals=True))
    _must_be_regular_tensor(sln.access_values())
    _must_be_read_only_tensor(sln.values)
    _must_be_read_only_tensor(sln.evals)
    testing.assert_shape_matches(sln.access_values(keep_evals=True), length)
    testing.assert_shape_matches(sln.access_evals(), 1)
    testing.assert_shape_matches(sln.values, length)
    testing.assert_shape_matches(sln.evals, 1)


@pytest.mark.parametrize(
    "problem_cls,eval_data_length",
    product(
        [DummyProblems.Batched, DummyProblems.MultiObj],
        [None, 0, 2, 10],
    ),
)
def test_set_evals(problem_cls: Callable, eval_data_length: Optional[int]):
    problem: et.Problem = problem_cls(eval_data_length=eval_data_length)
    num_objs = len(problem.senses)
    popsize = 5
    dtype = problem.eval_dtype
    tolerance = 1e-4

    # =============================================

    if num_objs == 1:
        # Make a new batch
        batch = problem.generate_batch(popsize)

        # Produce some arbitrary evaluation results
        f = torch.arange(popsize, dtype=dtype)

        # Generate the expected contents of the evals tensor
        e = torch.empty(batch.evals.shape, dtype=dtype)
        e[:] = float("nan")
        e[:, 0] = f

        # Set the evaluation results in the batch
        batch.set_evals(f)

        # Is the evals tensor looking as expected?
        testing.assert_allclose(batch.evals, e, atol=tolerance)

    # =============================================
    # (Re-)make the batch
    batch = problem.generate_batch(popsize)

    # Produce some arbitrary evaluation results
    f = torch.randn((popsize, num_objs), dtype=dtype)

    # Generate the expected contents of the evals tensor
    e = torch.empty(batch.evals.shape, dtype=dtype)
    e[:] = float("nan")
    e[:, :num_objs] = f

    # Set the evaluation results in the batch
    batch.set_evals(f)

    # Is the evals tensor looking as expected?
    testing.assert_allclose(batch.evals, e, atol=tolerance)

    # =============================================
    if (eval_data_length is not None) and (eval_data_length > 0):
        # Re-make the batch
        batch = problem.generate_batch(popsize)

        # Produce some arbitrary evaluation results
        f = torch.randn((popsize, num_objs), dtype=dtype)

        # Produce some arbitrary evaluation data
        d = torch.randn((popsize, eval_data_length), dtype=dtype)

        # Generate the expected contents of the evals tensor
        e = torch.empty(batch.evals.shape, dtype=dtype)
        e[:, :num_objs] = f
        e[:, num_objs:] = d

        # Set the evaluation results in the batch
        batch.set_evals(f, d)

        # Is the evals tensor looking as expected?
        testing.assert_allclose(batch.evals, e, atol=tolerance)

        # =============================================
        # Re-make the batch
        batch = problem.generate_batch(popsize)

        # Produce some arbitrary evaluation results and evaluation data
        f = torch.randn((popsize, num_objs + eval_data_length), dtype=dtype)

        # Set the evaluation results in the batch
        batch.set_evals(f)

        # Is the evals tensor looking as expected?
        testing.assert_allclose(batch.evals, f, atol=tolerance)

    # =============================================
    if num_objs >= 2:
        for two_dim in False, True:
            # Re-make the batch
            batch = problem.generate_batch(popsize)

            # Produce some evaluation results, deliberately in a wrong shape
            f = torch.arange(popsize, dtype=dtype)
            if two_dim:
                f = f.reshape(popsize, 1)

            # This must fail
            with pytest.raises(ValueError):
                batch.set_evals(f)

    # =============================================
    if (eval_data_length is not None) and eval_data_length >= 2:
        # Re-make the batch
        batch = problem.generate_batch(popsize)

        # Produce some evaluation results, deliberately in a wrong shape
        f = torch.randn((popsize, num_objs + 1), dtype=dtype)

        # This must fail
        with pytest.raises(ValueError):
            batch.set_evals(f)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the ObjectArray class, which is an array-like
data structure with an interface similar to PyTorch tensors,
but with an ability to store arbitrary type of data (not just numbers).
"""

from collections.abc import Sequence
from copy import deepcopy
from typing import Any, Iterable, Optional

import numpy as np
import torch

from .misc import Device, DType, Size, clone, is_integer, is_sequence


class ObjectArrayStorage:
    def __init__(self, source: "ObjectArray"):
        self._source = source

    def data_ptr(self) -> int:
        return id(self._source._objects)


class ObjectArray(Sequence):
    """
    An object container with an interface similar to PyTorch tensors.

    It is strictly one-dimensional, and supports advanced indexing and
    slicing operations supported by PyTorch tensors.

    An ObjectArray can store `None` values, strings, numbers, booleans,
    lists, sets, dictionaries, PyTorch tensors, and numpy arrays.

    When a container (such as a list, dictionary, set, is placed into an
    ObjectArray, an immutable clone of this container is first created, and
    then this newly created immutable clone gets stored within the
    ObjectArray. This behavior is to prevent accidental modification of the
    stored data.

    When a numeric array (such as a PyTorch tensor or a numpy array with a
    numeric dtype) is placed into an ObjectArray, the target ObjectArray
    first checks if the numeric array is read-only. If the numeric array
    is indeed read-only, then the array is put into the ObjectArray as it
    is. If the array is not read-only, then a read-only clone of the
    original numeric array is first created, and then this clone gets
    stored by the ObjectArray. This behavior has the following implications:
    (i) even when an ObjectArray is shared by multiple components of the
    program, the risk of accidental modification of the stored data through
    this shared ObjectArray is significantly reduced as the stored numeric
    arrays are read-only;
    (ii) although not recommended, one could still forcefully modify the
    numeric arrays stored by an ObjectArray by explicitly casting them as
    mutable arrays
    (in the case of a numpy array, one could forcefully set the WRITEABLE
    flag, and, in the case of a ReadOnlyTensor, one could forcefully cast it
    as a regular PyTorch tensor);
    (iii) if an already read-only array `x` is placed into an ObjectArray,
    but `x` shares its memory with a mutable array `y`, then the contents
    of the ObjectArray can be affected by modifying `y`.
    The implication (ii) is demonstrated as follows:

    ```python
    objs = ObjectArray(1)  # a single-element ObjectArray

    # Place a numpy array into objs:
    objs[0] = np.array([1, 2, 3], dtype=float)

    # At this point, objs[0] is a read-only numpy array.
    # objs[0] *= 2   # <- Not allowed

    # Possible but NOT recommended:
    objs.flags["WRITEABLE"] = True
    objs[0] *= 2
    ```

    The implication (iii) is demonstrated as follows:

    ```python
    objs = ObjectArray(1)  # a single-element ObjectArray

    # Make a new mutable numpy array
    y = np.array([1, 2, 3], dtype=float)

    # Make a read-only view to y:
    x = y[:]
    x.flags["WRITEABLE"] = False

    # Place x into objs.
    objs[0] = x

    # At this point, objs[0] is a read-only numpy array.
    # objs[0] *= 2   # <- Not allowed

    # During the operation of setting its 0-th item, the ObjectArray
    # `objs` did not clone `x` because `x` was already read-only.
    # However, the contents of `x` could actually be modified because
    # `x` shares its memory with the mutable array `y`.

    # Possible but NOT recommended:
    y *= 2    # This affects both x and objs!
    ```

    When a numpy array of dtype object is placed into an ObjectArray,
    a read-only ObjectArray copy of the original array will first be
    created, and then, this newly created ObjectArray will be stored
    by the outer ObjectArray.

    An ObjectArray itself has a read-only mode, so that, in addition to its
    stored data, the ObjectArray itself can be protected against undesired
    modifications.

    An interesting feature of PyTorch: if one slices a tensor A and the
    result is a new tensor B, and if B is sharing storage memory with A,
    then A.storage().data_ptr() and B.storage().data_ptr() will return
    the same pointer. This means, one can compare the storage pointers of
    A and B and see whether or not the two are sharing memory.
    ObjectArray was designed to have this exact behavior, so that one
    can understand if two ObjectArray instances are sharing memory.
    Note that NumPy does NOT have such a behavior. In more details,
    a NumPy array C and a NumPy array D could report different pointers
    even when D was created via a basic slicing operation on C.
    """

    def __init__(
        self,
        size: Optional[Size] = None,
        *,
        slice_of: Optional[tuple] = None,
    ):
        """
        `__init__(...)`: Instantiate a new ObjectArray.

        Args:
            size: Length of the ObjectArray. If this argument is present and
                is an integer `n`, then the resulting ObjectArray will be
                of length `n`, and will be filled with `None` values.
                This argument cannot be used together with the keyword
                argument `slice_of`.
            slice_of: Optionally a tuple in the form
                `(original_object_tensor, slice_info)`.
                When this argument is present, then the resulting ObjectArray
                will be a slice of the given `original_object_tensor` (which
                is expected as an ObjectArray instance). `slice_info` is
                either a `slice` instance, or a sequence of integers.
                The resulting ObjectArray might be a view of
                `original_object_tensor` (i.e. it might share its memory with
                `original_object_tensor`).
                This keyword argument cannot be used together with the
                argument `size`.
        """
        if size is not None and slice_of is not None:
            raise ValueError("Expected either `size` argument or `slice_of` argument, but got both.")
        elif size is None and slice_of is None:
            raise ValueError("Expected either `size` argument or `slice_of` argument, but got none.")
        elif size is not None:
            if not is_sequence(size):
                length = size
            elif isinstance(size, (np.ndarray, torch.Tensor)) and (size.ndim > 1):
                raise ValueError(f"Invalid size: {size}")
            else:
                [length] = size
            length = int(length)
            self._indices = torch.arange(length, dtype=torch.int64)
            self._objects = [None] * length
        elif slice_of is not None:
            source: ObjectArray

            source, slicing = slice_of

            if not isinstance(source, ObjectArray):
                raise TypeError(
                    f"`slice_of`: The first element was expected as an ObjectArray."
                    f" But it is of type {repr(type(source))}"
                )

            if isinstance(slicing, tuple) or is_integer(slicing):
                raise TypeError(f"Invalid slice: {slicing}")

            self._indices = source._indices[slicing]
            self._objects = source._objects

            if self._indices.storage().data_ptr() != source._indices.storage().data_ptr():
                self._objects = clone(self._objects)

        self._device = torch.device("cpu")
        self._read_only = False

    @property
    def shape(self) -> Size:
        """Shape of the ObjectArray, as a PyTorch Size tuple."""
        return self._indices.shape

    def size(self) -> Size:
        """
        Get the size of the ObjectArray, as a PyTorch Size tuple.

        Returns:
            The size (i.e. the shape) of the ObjectArray.
        """
        return self._indices.size()

    @property
    def ndim(self) -> int:
        """
        Number of dimensions handled by the ObjectArray.
        This is equivalent to getting the length of the size tuple.
        """
        return self._indices.ndim

    def dim(self) -> int:
        """
        Get the number of dimensions handled by the ObjectArray.
        This is equivalent to getting the length of the size tuple.

        Returns:
            The number of dimensions, as an integer.
        """
        return self._indices.dim()

    def numel(self) -> int:
        """
        Number of elements stored by the ObjectArray.

        Returns:
            The number of elements, as an integer.
        """
        return self._indices.numel()

    def repeat(self, *sizes) -> "ObjectArray":
        """
        Repeat the contents of this ObjectArray.

        For example, if we have an ObjectArray `objs` which stores
        `["hello", "world"]`, the following line:

            objs.repeat(3)

        will result in an ObjectArray which stores:

            `["hello", "world", "hello", "world", "hello", "world"]`

        Args:
            sizes: Although this argument is named `sizes` to be compatible
                with PyTorch, what is expected here is a single positional
                argument, as a single integer, or as a single-element
                tuple.
                The given integer (which can be the argument itself, or
                the integer within the given single-element tuple),
                specifies how many times the stored sequence will be
                repeated.
        Returns:
            A new ObjectArray which repeats the original one's values
        """

        if len(sizes) != 1:
            type_name = type(self).__name__
            raise ValueError(
                f"The `repeat(...)` method of {type_name} expects exactly one positional argument."
                f" This is because {type_name} supports only 1-dimensional storage."
                f" The received positional arguments are: {sizes}."
            )
        if isinstance(sizes, tuple):
            if len(sizes) == 1:
                sizes = sizes[0]
            else:
                type_name = type(self).__name__
                raise ValueError(
                    f"The `repeat(...)` method of {type_name} can accept a size tuple with only one element."
                    f" This is because {type_name} supports only 1-dimensional storage."
                    f" The received size tuple is: {sizes}."
                )
        num_repetitions = int(sizes[0])
        self_length = len(self)
        result = ObjectArray(num_repetitions * self_length)

        source_index = 0
        for result_index in range(len(result)):
            result[result_index] = self[source_index]
            source_index = (source_index + 1) % self_length

        return result

    @property
    def device(self) -> Device:
        """
        The device which stores the elements of the ObjectArray.
        In the case of ObjectArray, this property always returns
        the CPU device.

        Returns:
            The CPU device, as a torch.device object.
        """
        return self._device

    @property
    def dtype(self) -> DType:
        """
        The dtype of the elements stored by the ObjectArray.
        In the case of ObjectArray, the dtype is always `object`.
        """
        return object

    def __getitem__(self, i: Any) -> Any:
        if is_integer(i):
            index = int(self._indices[i])
            return self._objects[index]
        else:
            indices = self._indices[i]

            same_ptr = indices.storage().data_ptr() == self._indices.storage().data_ptr()

            result = ObjectArray(len(indices))

            if same_ptr:
                result._indices[:] = indices
                result._objects = self._objects
            else:
                result._objects = []
                for index in indices:
                    result._objects.append(self._objects[int(index)])

            result._read_only = self._read_only

            return result

    def __setitem__(self, i: Any, x: Any):
        from .immutable import as_immutable

        if self._read_only:
            raise ValueError("This ObjectArray is read-only, therefore, modification is not allowed.")
        if is_integer(i):
            index = int(self._indices[i])
            self._objects[index] = as_immutable(x)
        else:
            indices = self._indices[i]
            if not isinstance(x, Iterable):
                raise TypeError(f"Expected an iterable, but got {repr(x)}")
            if not hasattr(x, "__len__"):
                x = list(x)
            if len(x) != len(indices):
                raise TypeError(
                    f"The slicing operation refers to {len(indices)} elements."
                    f" However, the given objects sequence has {len(x)} elements."
                )
            for q, obj in enumerate(x):
                index = int(indices[q])
                self._objects[index] = as_immutable(obj)

    def __len__(self) -> int:
        return len(self._indices)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def clone(self, *, memo: Optional[dict] = None) -> "ObjectArray":
        """
        Get a deep copy of the ObjectArray.

        Note that the newly made deep copy will NOT be read-only,
        even if the original is.

        Returns:
            An non-read-only deep copy of the original ObjectArray.
        """
        if memo is None:
            memo = {}
        result = ObjectArray(len(self))
        for i in range(len(self)):
            result[i] = deepcopy(self[i], memo=memo)
        return result

    def get_read_only_view(self) -> "ObjectArray":
        """
        Get a read-only view of this ObjectArray.
        """
        result = self[:]
        result._read_only = True
        return result

    @property
    def is_read_only(self) -> bool:
        """
        True if this ObjectArray is read-only; False otherwise.
        """
        return self._read_only

    def __copy__(self) -> "ObjectArray":
        return self.clone()

    def __deepcopy__(self, memo: Optional[dict]) -> "ObjectArray":
        return self.clone(memo=memo)

    def __getstate__(self):
        return self.clone().__dict__

    def storage(self) -> ObjectArrayStorage:
        return ObjectArrayStorage(self)

    def _to_string(self) -> str:
        inside = []
        for ind in self._indices:
            i = int(ind)
            inside.append(self._objects[i])
        type_name = type(self).__name__
        details = [
            "elements: " + repr(inside),
            "ptr: " + repr(self.storage().data_ptr()),
        ]
        if self.is_read_only:
            details.append("is_read_only: " + repr(self.is_read_only))
        details = ", ".join(details)
        return f"<{type_name}, {details}>"

    def __repr__(self) -> str:
        return self._to_string()

    def __str__(self) -> str:
        return self._to_string()

    def numpy(self) -> np.ndarray:
        """
        Convert this ObjectArray to a numpy array.

        The resulting numpy array will have its dtype set as `object`.
        This new array itself and its contents will be mutable (those
        mutable objects being the copies of their immutable sources).

        Returns:
            The numpy counterpart of this ObjectArray.
        """
        from .immutable import mutable_copy

        n = len(self)
        result = np.empty(n, dtype=object)
        for i, item in enumerate(self):
            if isinstance(item, ObjectArray):
                result[i] = item.numpy()
            else:
                result[i] = mutable_copy(item)

        return result

    @staticmethod
    def from_numpy(ndarray: np.ndarray) -> "ObjectArray":
        """
        Convert a numpy array of dtype `object` to an `ObjectArray`.

        Args:
            The numpy array that will be converted to `ObjectArray`.
        Returns:
            The ObjectArray counterpart of the given numpy array.
        """
        if isinstance(ndarray, np.ndarray):
            if ndarray.dtype == np.dtype(object):
                n = len(ndarray)
                result = ObjectArray(n)
                for i, element in enumerate(ndarray):
                    result[i] = element
                return result
            else:
                raise ValueError(
                    f"The dtype of the given array was expected as `object`."
                    f" However, the dtype was encountered as {ndarray.dtype}."
                )
        else:
            raise TypeError(f"Expected a `numpy.ndarray` instance, but received an object of type {type(ndarray)}.")



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for parsing string representations of neural net policies
"""

import ast
from pprint import pprint
from typing import Any, Callable, Optional, Type

import torch
from torch import nn

from . import layers


class NetParsingError(Exception):
    """
    Representation of a parsing error
    """

    def __init__(
        self,
        message: str,
        lineno: Optional[int] = None,
        col_offset: Optional[int] = None,
        original_error: Optional[Exception] = None,
    ):
        """
        `__init__(...)`: Initialize the NetParsingError.

        Args:
            message: Error message, as string.
            lineno: Erroneous line number in the string representation of the
                neural network structure.
            col_offset: Erroneous column number in the string representation
                of the neural network structure.
            original_error: If another error caused this parsing error,
                that original error can be attached to this `NetParsingError`
                instance via this argument.
        """
        super().__init__()
        self.message = message
        self.lineno = lineno
        self.col_offset = col_offset
        self.original_error = original_error

    def _to_string(self) -> str:
        parts = []

        parts.append(type(self).__name__)

        if self.lineno is not None:
            parts.append(" at line(")
            parts.append(str(self.lineno - 1))
            parts.append(")")

        if self.col_offset is not None:
            parts.append(" at column(")
            parts.append(str(self.col_offset + 1))
            parts.append(")")

        parts.append(": ")
        parts.append(self.message)

        return "".join(parts)

    def __str__(self) -> str:
        return self._to_string()

    def __repr__(self) -> str:
        return self._to_string()


def submodules(a: nn.Module) -> list:
    if isinstance(a, nn.Sequential):
        return [module for i, module in enumerate(a.modules()) if i >= 1]
    else:
        return [a]


def concat_modules(a: nn.Module, b: nn.Module) -> nn.Module:
    all_modules = submodules(a) + submodules(b)
    return nn.Sequential(*all_modules)


def _get_nn_module(name: str) -> Type:
    if hasattr(layers, name):
        return getattr(layers, name)
    elif hasattr(nn, name):
        return getattr(nn, name)
    else:
        raise NetParsingError(f"Unrecognized module class: {repr(name)}")


def _eval_with_constants(node: ast.expr, constants: dict) -> Any:
    def fail(msg, erroneous_node=None, original_error=None):
        nonlocal node

        if (
            erroneous_node is None
            or (not hasattr(erroneous_node, "lineno"))
            or (not hasattr(erroneous_node, "col_offset"))
        ):
            erroneous_node = node

        raise NetParsingError(msg, erroneous_node.lineno, erroneous_node.col_offset, original_error=original_error)

    def literal_eval(subnode: ast.expr):
        err = None
        result = None

        try:
            result = ast.literal_eval(subnode)
        except Exception as ex:
            err = ex

        if err is not None:
            fail(f"When trying to parse expression, encountered: {repr(err)}", subnode, original_error=err)

        return result

    def get_constant(name):
        nonlocal constants

        if name not in constants.keys():
            fail(f"Unknown constant: {name}. Available constants: {repr(list(constants.keys()))}")

        return constants[name]

    def get_from_constant(name, index):
        cnst = get_constant(name)

        err = None
        result = None

        try:
            result = cnst[index]
        except Exception as ex:
            err = ex

        if err is not None:
            fail(
                f"When applying the indexing operation on the constant {name}, encountered: {repr(err)}",
                original_error=err,
            )

        return result

    if isinstance(node, ast.Name):
        return get_constant(node.id)
    elif isinstance(node, ast.Subscript):
        name = node.value
        if not isinstance(name, ast.Name):
            fail(
                "Expression which was expected to express a simple indexing over a constant"
                " is either too complex, or is unrecognized.",
                name,
            )
        name = name.id

        index = node.slice
        if not isinstance(index, ast.Index):
            fail(f"Expected a simple indexing operation, but got a {type(index).__name__}.", index)
        index = literal_eval(index.value)

        return get_from_constant(name, index)
    else:
        return literal_eval(node)


def _process_call_expr(node: ast.Call, constants: dict) -> nn.Module:
    args = [_eval_with_constants(arg, constants=constants) for arg in node.args]

    kwargs = {}
    for kw in node.keywords:
        kwname = kw.arg
        kwvalue = _eval_with_constants(kw.value, constants=constants)
        kwargs[kwname] = kwvalue

    name = node.func.id

    return _get_nn_module(name)(*args, **kwargs)


def _process_rshift_expr(node: ast.BinOp, constants: dict) -> nn.Module:
    if not isinstance(node.op, ast.RShift):
        raise NetParsingError("Binary operators other than '>>' are not recognized.", node.lineno, node.col_offset)
    left_module = _process_expr(node.left, constants=constants)
    right_module = _process_expr(node.right, constants=constants)
    return concat_modules(left_module, right_module)


def _process_expr(node: ast.expr, constants: dict) -> nn.Module:
    if isinstance(node, ast.Call):
        return _process_call_expr(node, constants=constants)
    elif isinstance(node, ast.BinOp):
        return _process_rshift_expr(node, constants=constants)
    else:
        raise NetParsingError(f"Unrecognized expression of type {type(node)}", node.lineno, node.col_offset)


def str_to_net(s: str, **constants) -> nn.Module:
    """
    Read a string representation of a neural net structure,
    and return a `torch.nn.Module` instance out of it.

    Let us imagine that one wants to describe the following
    neural network structure:

    ```python
    from torch import nn

    net = nn.Sequential(
        nn.Linear(8, 16),
        nn.Tanh(),
        nn.Linear(16, 4, bias=False),
        nn.ReLU()
    )
    ```

    By using `str_to_net(...)` one can construct the same
    module via:

    ```python
    from evotorch.neuroevolution.net import str_to_net

    net = str_to_net(
        'Linear(8, 16) >> Tanh() >> Linear(16, 4, bias=False) >> ReLU()'
    )
    ```

    The string can also be multi-line:

    ```python
    net = str_to_net(
        '''
        Linear(8, 16)
        >> Tanh()
        >> Linear(16, 4, bias=False)
        >> ReLU()
        '''
    )
    ```

    One can also define constants for using them in strings:

    ```python
    net = str_to_net(
        '''
        Linear(input_size, hidden_size)
        >> Tanh()
        >> Linear(hidden_size, output_size, bias=False)
        >> ReLU()
        ''',
        input_size=8,
        hidden_size=16,
        output_size=4
    )
    ```

    In the neural net structure string, when one refers to a module type,
    say, `Linear`, first the name `Linear` is searched for in the namespace
    `evotorch.neuroevolution.net.layers`, and then in the namespace `torch.nn`.
    In the case of `Linear`, the searched name exists in `torch.nn`,
    and therefore, the layer type to be instantiated is accepted as
    `torch.nn.Linear`.
    Instead of `Linear`, if one had used the name, say,
    `StructuredControlNet`, then, the layer type to be instantiated
    would be `evotorch.neuroevolution.net.layers.StructuredControlNet`.

    **Notes regarding usage with `evotorch.neuroevolution.GymNE`:**
    While instantiating a `GymNE`, one can specify a neural net
    structure string as the policy. Therefore, while filling the policy
    string for a `GymNE`, all these rules mentioned above apply.
    Additionally, while using `str_to_net(...)` internally,
    `GymNE` defines these extra constants:
    `obs_length` (length of the observation vector),
    `act_length` (length of the action vector for continuous-action
    environments, or number of actions for discrete-action
    environments), and `obs_shape` (shape of the observation as a tuple,
    assuming that the observation space is of type `gym.spaces.Box`,
    usable within the string like `obs_shape[0]`, `obs_shape[1]`, etc.,
    or simply `obs_shape` to refer to the entire tuple).

    Therefore, while using with `GymNE`, one can define a
    single-hidden-layered policy via this string:

    ```python
    'Linear(obs_length, 16) >> Tanh() >> Linear(16, act_length) >> Tanh()'
    ```

    (where one might choose to omit the last `Tanh()` as `GymNE`
    will clip the output of the final layer to conform with the
    action boundaries of the environment, which one might think as a
    type of hard-tanh anyway).

    Args:
        s: The string which expresses the neural net structure.
    Returns:
        The PyTorch module of the specified structure.
    """
    s = f"(\n{s}\n)"
    return _process_expr(ast.parse(s, mode="eval").body, constants=constants)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
from copy import copy, deepcopy
from typing import Any, Callable, Iterable, Mapping, Optional, Union

import numpy as np
import torch

_torch_older_than_1_12 = False


def _check_torch_version():
    # Determine the PyTorch version without failing when __version__ has an unexpected value.
    # Perhaps such unexpected values could be encountered when using a custom/modified version
    # of PyTorch, and we do not wish this module to fail in those scenarios.
    global _torch_older_than_1_12

    if hasattr(torch, "__version__"):
        ver = torch.__version__
        if isinstance(ver, str):
            vers = ver.split(".")
            with contextlib.suppress(Exception):
                a = int(vers[0])
                b = int(vers[1])
                if (a <= 1) and (b < 12):
                    _torch_older_than_1_12 = True


_check_torch_version()


class ReadOnlyTensor(torch.Tensor):
    """
    A special type of tensor which is read-only.

    This is a subclass of `torch.Tensor` which explicitly disallows
    operations that would cause in-place modifications.

    Since ReadOnlyTensor if a subclass of `torch.Tensor`, most
    non-destructive PyTorch operations are on this tensor are supported.

    Cloning a ReadOnlyTensor using the `clone()` method or Python's
    `deepcopy(...)` function results in a regular PyTorch tensor.

    Reshaping or slicing operations might return a ReadOnlyTensor if the
    result ends up being a view of the original ReadOnlyTensor; otherwise,
    the returned tensor is a regular `torch.Tensor`.
    """

    def __getattribute__(self, attribute_name: str) -> Any:
        if (
            isinstance(attribute_name, str)
            and attribute_name.endswith("_")
            and (not ((attribute_name.startswith("__")) and (attribute_name.endswith("__"))))
        ):
            raise AttributeError(
                f"A ReadOnlyTensor explicitly disables all members whose names end with '_'."
                f" Cannot access member {repr(attribute_name)}."
            )
        else:
            return super().__getattribute__(attribute_name)

    def __cannot_modify(self, *ignore, **ignore_too):
        raise TypeError("The contents of a ReadOnlyTensor cannot be modified")

    __setitem__ = __cannot_modify
    __iadd__ = __cannot_modify
    __iand__ = __cannot_modify
    __idiv__ = __cannot_modify
    __ifloordiv__ = __cannot_modify
    __ilshift__ = __cannot_modify
    __imatmul__ = __cannot_modify
    __imod__ = __cannot_modify
    __imul__ = __cannot_modify
    __ior__ = __cannot_modify
    __ipow__ = __cannot_modify
    __irshift__ = __cannot_modify
    __isub__ = __cannot_modify
    __itruediv__ = __cannot_modify
    __ixor__ = __cannot_modify

    if _torch_older_than_1_12:
        # Define __str__ and __repr__ for when using PyTorch 1.11 or older.
        # With PyTorch 1.12, overriding __str__ and __repr__ are not necessary.
        def __to_string(self) -> str:
            s = super().__repr__()
            if "\n" not in s:
                return f"ReadOnlyTensor({super().__repr__()})"
            else:
                indenter = " " * 4
                s = (indenter + s.replace("\n", "\n" + indenter)).rstrip()
                return f"ReadOnlyTensor(\n{s}\n)"

        __str__ = __to_string
        __repr__ = __to_string

    def clone(self) -> torch.Tensor:
        return super().clone().as_subclass(torch.Tensor)

    def __mutable_if_independent(self, other: torch.Tensor) -> torch.Tensor:
        self_ptr = self.storage().data_ptr()
        other_ptr = other.storage().data_ptr()
        if self_ptr != other_ptr:
            other = other.as_subclass(torch.Tensor)
        return other

    def __getitem__(self, index_or_slice) -> torch.Tensor:
        result = super().__getitem__(index_or_slice)
        return self.__mutable_if_independent(result)

    def reshape(self, *args, **kwargs) -> torch.Tensor:
        result = super().reshape(*args, **kwargs)
        return self.__mutable_if_independent(result)

    def numpy(self) -> np.ndarray:
        arr: np.ndarray = torch.Tensor.numpy(self)
        arr.flags["WRITEABLE"] = False
        return arr

    def __array__(self, *args, **kwargs) -> np.ndarray:
        arr: np.ndarray = super().__array__(*args, **kwargs)
        arr.flags["WRITEABLE"] = False
        return arr

    # def __copy__(self):
    #    return ReadOnlyTensor(copy(self.as_subclass(torch.Tensor)))

    # def __deepcopy__(self, memo):
    #    return deepcopy(self.as_subclass(torch.Tensor), memo)

    @classmethod
    def __torch_function__(cls, func: Callable, types: Iterable, args: tuple = (), kwargs: Optional[Mapping] = None):
        if (kwargs is not None) and ("out" in kwargs):
            if isinstance(kwargs["out"], ReadOnlyTensor):
                raise TypeError(
                    f"The `out` keyword argument passed to {func} is a ReadOnlyTensor."
                    f" A ReadOnlyTensor explicitly fails when referenced via the `out` keyword argument of any torch"
                    f" function."
                    f" This restriction is for making sure that the torch operations which could normally do in-place"
                    f" modifications do not operate on ReadOnlyTensor instances."
                )
        return super().__torch_function__(func, types, args, kwargs)


def _device_and_dtype_kwargs(*, dtype: Optional[torch.dtype], device: Optional[Union[str, torch.device]]) -> dict:
    result = {}
    if dtype is not None:
        result["dtype"] = dtype
    if device is not None:
        result["device"] = device
    return result


def read_only_tensor(
    x: Any, *, dtype: Optional[torch.dtype] = None, device: Optional[Union[str, torch.device]] = None
) -> Iterable:
    """
    Make a ReadOnlyTensor from the given object.

    The provided object can be a scalar, or an Iterable of numeric data,
    or an ObjectArray.

    This function can be thought as the read-only counterpart of PyTorch's
    `torch.tensor(...)` function.

    Args:
        x: The object from which the new ReadOnlyTensor will be made.
        dtype: The dtype of the new ReadOnlyTensor (e.g. torch.float32).
        device: The device in which the ReadOnlyTensor will be stored
            (e.g. "cpu").
    Returns:
        The new read-only tensor.
    """
    from .objectarray import ObjectArray

    kwargs = _device_and_dtype_kwargs(dtype=dtype, device=device)
    if isinstance(x, ObjectArray):
        if len(kwargs) != 0:
            raise ValueError(
                f"read_only_tensor(...): when making a read-only tensor from an ObjectArray,"
                f" the arguments `dtype` and `device` were not expected."
                f" However, the received keyword arguments are: {kwargs}."
            )
        return x.get_read_only_view()
    else:
        return torch.as_tensor(x, **kwargs).as_subclass(ReadOnlyTensor)


def as_read_only_tensor(
    x: Any, *, dtype: Optional[torch.dtype] = None, device: Optional[Union[str, torch.device]] = None
) -> Iterable:
    """
    Convert the given object to a ReadOnlyTensor.

    The provided object can be a scalar, or an Iterable of numeric data,
    or an ObjectArray.

    This function can be thought as the read-only counterpart of PyTorch's
    `torch.as_tensor(...)` function.

    Args:
        x: The object to be converted to a ReadOnlyTensor.
        dtype: The dtype of the new ReadOnlyTensor (e.g. torch.float32).
            If this argument is not specified, dtype will be inferred from `x`.
            For example, if `x` is a PyTorch tensor or a numpy array, its
            existing dtype will be kept.
        device: The device in which the ReadOnlyTensor will be stored
            (e.g. "cpu").
            If this argument is not specified, the device which is storing
            the original `x` will be re-used.
    Returns:
        The read-only counterpart of the provided object.
    """
    from .objectarray import ObjectArray

    kwargs = _device_and_dtype_kwargs(dtype=dtype, device=device)
    if isinstance(x, ObjectArray):
        if len(kwargs) != 0:
            raise ValueError(
                f"read_only_tensor(...): when making a read-only tensor from an ObjectArray,"
                f" the arguments `dtype` and `device` were not expected."
                f" However, the received keyword arguments are: {kwargs}."
            )
        return x.get_read_only_view()
    else:
        return torch.as_tensor(x, **kwargs).as_subclass(ReadOnlyTensor)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from copy import copy
from typing import Any, Callable, Iterable, Optional, Union

import torch

from .tools import Device, DType, cast_tensors_in_container, device_of_container, dtype_of_container
from .tools import multiply_rows_by_scalars as dot
from .tools import rowwise_sum as total
from .tools import to_torch_dtype
from .tools.ranking import rank
from .tools.tensormaker import TensorMakerMixin


class Distribution(TensorMakerMixin):
    """
    Base class for any search distribution.
    """

    MANDATORY_PARAMETERS = set()
    OPTIONAL_PARAMETERS = set()

    def __init__(
        self, *, solution_length: int, parameters: dict, dtype: Optional[DType] = None, device: Optional[Device] = None
    ):
        """
        `__init__(...)`: Initialize the Distribution.

        It is expected that one of these two conditions is met:
        (i) the inheriting search distribution class does not implement its
        own `__init__(...)` method; or
        (ii) the inheriting search distribution class has its own
        `__init__(...)` method, and calls `Distribution.__init__(...)`
        from there, during its initialization phase.

        Args:
            solution_length: Expected as an integer, this argument represents
                the solution length.
            parameters: Expected as a dictionary, this argument stores
                the parameters of the search distribution.
                For example, for a Gaussian distribution where `mu`
                represents the mean, and `sigma` represents the coverage
                area, this dictionary would have the keys "mu" and "sigma",
                and each of these keys would map to a PyTorch tensor.
            dtype: The dtype of the search distribution (e.g. torch.float32).
            device: The device of the search distribution (e.g. "cpu").
        """
        self.__solution_length: int = int(solution_length)

        self.__parameters: dict
        self.__dtype: torch.dtype
        self.__device: torch.device

        self.__check_correctness(parameters)

        cast_kwargs = {}
        if dtype is not None:
            cast_kwargs["dtype"] = to_torch_dtype(dtype)
        if device is not None:
            cast_kwargs["device"] = torch.device(device)

        if len(cast_kwargs) == 0:
            self.__parameters = copy(parameters)
        else:
            self.__parameters = cast_tensors_in_container(parameters, **cast_kwargs)

        self.__dtype = cast_kwargs.get("dtype", dtype_of_container(parameters))
        self.__device = cast_kwargs.get("device", device_of_container(parameters))

    def __check_correctness(self, parameters: dict):
        found_mandatory = 0
        for param_name in parameters.keys():
            if param_name in self.MANDATORY_PARAMETERS:
                found_mandatory += 1
            elif param_name in self.OPTIONAL_PARAMETERS:
                pass  # nothing to do
            else:
                raise ValueError(f"Unrecognized parameter: {repr(param_name)}")
        if found_mandatory < len(self.MANDATORY_PARAMETERS):
            raise ValueError(
                f"Not all mandatory parameters of this Distribution were specified."
                f" Mandatory parameters of this distribution: {self.MANDATORY_PARAMETERS};"
                f" optional parameters of this distribution: {self.OPTIONAL_PARAMETERS};"
                f" encountered parameters: {set(parameters.keys())}."
            )

    def to(self, device: Device) -> "Distribution":
        """
        Bring the Distribution onto a computational device.

        If the given device is already the device of this Distribution,
        then the Distribution itself will be returned.
        If the given device is different than the device of this
        Distribution, a copy of this Distribution on the given device
        will be created and returned.

        Args:
            device: The computation device onto which the Distribution
                will be brought.
        Returns:
            The Distribution on the target device.
        """
        if torch.device(self.device) == torch.device(device):
            return self
        else:
            cls = self.__class__
            return cls(solution_length=self.solution_length, parameters=self.parameters, device=device)

    def _fill(self, out: torch.Tensor, *, generator: Optional[torch.Generator] = None):
        """
        Fill the given tensor with samples from this search distribution.

        It is expected that the inheriting search distribution class
        has its own implementation for this method.

        Args:
            out: The PyTorch tensor that will be filled with the samples.
                This tensor is expected as 2-dimensional with its number
                of columns equal to the solution length declared by this
                distribution.
            generator: Optionally a PyTorch generator, to be used for
                sampling. None means that the global generator of PyTorch
                is to be used.
        """
        raise NotImplementedError

    def sample(
        self,
        num_solutions: Optional[int] = None,
        *,
        out: Optional[torch.Tensor] = None,
        generator: Any = None,
    ) -> torch.Tensor:
        """
        Sample solutions from this search distribution.

        Args:
            num_solutions: How many solutions will be sampled.
                If this argument is given as an integer and the argument
                `out` is left as None, then a new PyTorch tensor, filled
                with the samples from this distribution, will be generated
                and returned. The number of rows of this new tensor will
                be equal to the given `num_solutions`.
                If the argument `num_solutions` is provided as an integer,
                then the argument `out` is expected as None.
            out: The PyTorch tensor that will be filled with the samples
                of this distribution. This tensor is expected as a
                2-dimensional tensor with its number of columns equal to
                the solution length declared by this distribution.
                If the argument `out` is provided as a tensor, then the
                argument `num_solutions` is expected as None.
            generator: Optionally a PyTorch generator or any object which
                has a `generator` attribute (e.g. a Problem instance).
                If left as None, the global generator of PyTorch will be
                used.
        Returns:
            A 2-dimensional PyTorch tensor which stores the sampled solutions.
        """
        if (num_solutions is not None) and (out is not None):
            raise ValueError(
                "Received both `num_solutions` and `out` with values other than None."
                "Please provide only one of these arguments with a value other than None, not both of them."
            )
        elif (num_solutions is not None) and (out is None):
            num_solutions = int(num_solutions)
            out = self.make_empty(num_solutions=num_solutions)
        elif (num_solutions is None) and (out is not None):
            if out.ndim != 2:
                raise ValueError(
                    f"The `sample(...)` method can fill only 2-dimensional tensors."
                    f" However, the provided `out` tensor has {out.ndim} dimensions, its shape being {out.shape}."
                )
            _, num_cols = out.shape
            if num_cols != self.solution_length:
                raise ValueError(
                    f"The solution length declared by this distribution is {self.solution_length}."
                    f" However, the provided `out` tensor has {num_cols} columns."
                    f" The `sample(...)` method can only work with tensors whose number of columns are equal"
                    f" to the declared solution length."
                )
        else:
            raise ValueError(
                "Received both `num_solutions` and `out` as None."
                "Please provide one of these arguments with a value other than None."
            )
        self._fill(out, generator=generator)
        return out

    def _compute_gradients(self, samples: torch.Tensor, weights: torch.Tensor, ranking_used: Optional[str]) -> dict:
        """
        Compute the gradients out of the samples (sampled solutions)
        and weights (i.e. weights or ranks of the solutions, better
        solutions having numerically higher weights).

        It is expected that the inheriting class implements this method.

        Args:
            samples: The sampled solutions, as a 2-dimensional tensor.
            weights: Solution weights, as a 1-dimensional tensor of length
                `n`, `n` being the number of sampled solutions.
            ranking_used: Ranking that was used to obtain the weights.
        Returns:
            The gradient(s) in a dictionary.
        """
        raise NotImplementedError

    def compute_gradients(
        self,
        samples: torch.Tensor,
        fitnesses: torch.Tensor,
        *,
        objective_sense: str,
        ranking_method: Optional[str] = None,
    ) -> dict:
        """
        Compute and return gradients.

        Args:
            samples: The solutions that were sampled from this Distribution.
                The tensor passed via this argument is expected to have
                the same dtype and device with this Distribution.
            fitnesses: The evaluation results of the sampled solutions.
                If fitnesses are given with a different dtype (maybe because
                the eval_dtype of the Problem object is different than its
                decision variable dtype), then this method will first
                create an internal copy of the fitnesses with the correct
                dtype, and then will use those copied fitnesses for
                computing the gradients.
            objective_sense: The objective sense, expected as "min" or "max".
                In the case of "min", lower fitness values will be regarded
                as better (therefore, in this case, one can alternatively
                refer to fitnesses as 'unfitnesses' or 'solution costs').
                In the case of "max", higher fitness values will be regarded
                as better.
            ranking_method: The ranking method to be used.
                Can be: "linear" (where ranks linearly go from 0 to 1);
                "centered" (where ranks linearly go from -0.5 to +0.5);
                "normalized" (where the standard-normalized fitnesses
                serve as ranks); or "raw" (where the fitnesses themselves
                serve as ranks).
                The default is "raw".
        Returns:
            A dictionary which contains the gradient for each parameter of the
            distribution.
        """
        if objective_sense == "max":
            higher_is_better = True
        elif objective_sense == "min":
            higher_is_better = False
        else:
            raise ValueError(
                f'`objective_sense` was expected as "min" or as "max".'
                f" However, it was encountered as {repr(objective_sense)}."
            )

        if ranking_method is None:
            ranking_method = "raw"

        # Make sure that the fitnesses are in the correct dtype
        fitnesses = torch.as_tensor(fitnesses, dtype=self.dtype)

        [num_samples, _] = samples.shape
        [num_fitnesses] = fitnesses.shape
        if num_samples != num_fitnesses:
            raise ValueError(
                f"The number of samples and the number of fitnesses do not match:" f" {num_samples} != {num_fitnesses}."
            )

        weights = rank(fitnesses, ranking_method=ranking_method, higher_is_better=higher_is_better)
        return self._compute_gradients(samples, weights, ranking_method)

    def update_parameters(
        self,
        gradients: dict,
        *,
        learning_rates: Optional[dict] = None,
        optimizers: Optional[dict] = None,
    ) -> "Distribution":
        """
        Do an update on the distribution by following the given gradients.

        It is expected that the inheriting class has its own implementation
        for this method.

        Args:
            gradients: Gradients, as a dictionary, which will be used for
                computing the necessary updates.
            learning_rates: A dictionary which contains learning rates
                for parameters that will be updated using a learning rate
                coefficient.
            optimizers: A dictionary which contains optimizer objects
                for parameters that will be updated using an adaptive
                optimizer.
        Returns:
            The updated copy of the distribution.
        """
        raise NotImplementedError

    def modified_copy(
        self, *, dtype: Optional[DType] = None, device: Optional[Device] = None, **parameters
    ) -> "Distribution":
        """
        Return a modified copy of this distribution.

        Args:
            dtype: The new dtype of the distribution.
            device: The new device of the distribution.
            parameters: Expected in the form of extra keyword arguments.
                Each of these keyword arguments will cause the new distribution
                to have a modified value for the specified parameter.
        Returns:
            The modified copy of the distribution.
        """
        cls = self.__class__
        if device is None:
            device = self.device
        if dtype is None:
            dtype = self.dtype

        new_parameters = copy(self.parameters)
        new_parameters.update(parameters)
        return cls(parameters=new_parameters, dtype=dtype, device=device)

    def relative_entropy(dist_0: "Distribution", dist_1: "Distribution") -> float:
        raise NotImplementedError

    @property
    def solution_length(self) -> int:
        return self.__solution_length

    @property
    def device(self) -> torch.device:
        return self.__device

    @property
    def dtype(self) -> torch.dtype:
        return self.__dtype

    @property
    def parameters(self) -> dict:
        return self.__parameters

    def _follow_gradient(
        self,
        param_name: str,
        x: torch.Tensor,
        *,
        learning_rates: Optional[dict] = None,
        optimizers: Optional[dict] = None,
    ) -> torch.Tensor:
        x = torch.as_tensor(x, dtype=self.dtype, device=self.device)
        learning_rate, optimizer = self._get_learning_rate_and_optimizer(param_name, learning_rates, optimizers)
        if (learning_rate is None) and (optimizer is None):
            return x
        elif (learning_rate is not None) and (optimizer is None):
            return learning_rate * x
        elif (learning_rate is None) and (optimizer is not None):
            return optimizer.ascent(x)
        else:
            raise ValueError(
                "Encountered both `learning_rate` and `optimizer` as values other than None."
                " This method can only work if both of them are None or only one of them is not None."
            )

    @staticmethod
    def _get_learning_rate_and_optimizer(
        param_name: str, learning_rates: Optional[dict], optimizers: Optional[dict]
    ) -> tuple:
        if learning_rates is None:
            learning_rates = {}
        if optimizers is None:
            optimizers = {}
        return learning_rates.get(param_name, None), optimizers.get(param_name, None)


class SeparableGaussian(Distribution):
    """Separable Multivariate Gaussian, as used by PGPE"""

    MANDATORY_PARAMETERS = {"mu", "sigma"}
    OPTIONAL_PARAMETERS = {"divide_mu_grad_by", "divide_sigma_grad_by", "parenthood_ratio"}

    def __init__(
        self,
        parameters: dict,
        *,
        solution_length: Optional[int] = None,
        device: Optional[Device] = None,
        dtype: Optional[DType] = None,
    ):
        [mu_length] = parameters["mu"].shape
        [sigma_length] = parameters["sigma"].shape

        if solution_length is None:
            solution_length = mu_length
        else:
            if solution_length != mu_length:
                raise ValueError(
                    f"The argument `solution_length` does not match the length of `mu` provided in `parameters`."
                    f" solution_length={solution_length},"
                    f' parameters["mu"]={mu_length}.'
                )

        if mu_length != sigma_length:
            raise ValueError(
                f"The tensors `mu` and `sigma` provided within `parameters` have mismatching lengths."
                f' parameters["mu"]={mu_length},'
                f' parameters["sigma"]={sigma_length}.'
            )

        super().__init__(
            solution_length=solution_length,
            parameters=parameters,
            device=device,
            dtype=dtype,
        )

    @property
    def mu(self) -> torch.Tensor:
        return self.parameters["mu"]

    @mu.setter
    def mu(self, new_mu: Iterable):
        self.parameters["mu"] = torch.as_tensor(new_mu, dtype=self.dtype, device=self.device)

    @property
    def sigma(self) -> torch.Tensor:
        return self.parameters["sigma"]

    @sigma.setter
    def sigma(self, new_sigma: Iterable):
        self.parameters["sigma"] = torch.as_tensor(new_sigma, dtype=self.dtype, device=self.device)

    def _fill(self, out: torch.Tensor, *, generator: Optional[torch.Generator] = None):
        self.make_gaussian(out=out, center=self.mu, stdev=self.sigma, generator=generator)

    def _divide_grad(self, param_name: str, grad: torch.Tensor, weights: torch.Tensor) -> torch.Tensor:
        option = f"divide_{param_name}_grad_by"
        if option in self.parameters:
            div_by_what = self.parameters[option]
            if div_by_what == "num_solutions":
                [num_solutions] = weights.shape
                grad = grad / num_solutions
            elif div_by_what == "num_directions":
                [num_solutions] = weights.shape
                num_directions = num_solutions // 2
                grad = grad / num_directions
            elif div_by_what == "total_weight":
                total_weight = torch.sum(torch.abs(weights))
                grad = grad / total_weight
            elif div_by_what == "weight_stdev":
                weight_stdev = torch.std(weights)
                grad = grad / weight_stdev
            else:
                raise ValueError(f"The parameter {option} has an unrecognized value: {div_by_what}")
        return grad

    def _compute_gradients_via_parenthood_ratio(self, samples: torch.Tensor, weights: torch.Tensor) -> dict:
        [num_samples, _] = samples.shape
        num_elites = math.floor(num_samples * self.parameters["parenthood_ratio"])
        elite_indices = weights.argsort(descending=True)[:num_elites]
        elites = samples[elite_indices, :]
        return {
            "mu": torch.mean(elites, dim=0) - self.parameters["mu"],
            "sigma": torch.std(elites, dim=0) - self.parameters["sigma"],
        }

    def _compute_gradients(self, samples: torch.Tensor, weights: torch.Tensor, ranking_used: Optional[str]) -> dict:
        if "parenthood_ratio" in self.parameters:
            return self._compute_gradients_via_parenthood_ratio(samples, weights)
        else:
            mu = self.mu
            sigma = self.sigma

            # Compute the scaled noises, that is, the noise vectors which
            # were used for generating the solutions
            # (solution = scaled_noise + center)
            scaled_noises = samples - mu

            # Make sure that the weights (utilities) are 0-centered
            # (Otherwise the formulations would have to consider a bias term)
            if ranking_used not in ("centered", "normalized"):
                weights = weights - torch.mean(weights)

            mu_grad = self._divide_grad(
                "mu",
                total(dot(weights, scaled_noises)),
                weights,
            )
            sigma_grad = self._divide_grad(
                "sigma",
                total(dot(weights, ((scaled_noises**2) - (sigma**2)) / sigma)),
                weights,
            )

            return {
                "mu": mu_grad,
                "sigma": sigma_grad,
            }

    def update_parameters(
        self,
        gradients: dict,
        *,
        learning_rates: Optional[dict] = None,
        optimizers: Optional[dict] = None,
    ) -> "SeparableGaussian":
        mu_grad = gradients["mu"]
        sigma_grad = gradients["sigma"]

        new_mu = self.mu + self._follow_gradient("mu", mu_grad, learning_rates=learning_rates, optimizers=optimizers)
        new_sigma = self.sigma + self._follow_gradient(
            "sigma", sigma_grad, learning_rates=learning_rates, optimizers=optimizers
        )

        return self.modified_copy(mu=new_mu, sigma=new_sigma)

    def relative_entropy(dist_0: "SeparableGaussian", dist_1: "SeparableGaussian") -> float:
        mu_0 = dist_0.parameters["mu"]
        mu_1 = dist_1.parameters["mu"]
        sigma_0 = dist_0.parameters["sigma"]
        sigma_1 = dist_1.parameters["sigma"]
        cov_0 = sigma_0.pow(2.0)
        cov_1 = sigma_1.pow(2.0)

        mu_delta = mu_1 - mu_0

        trace_cov = torch.sum(cov_0 / cov_1)
        k = dist_0.solution_length
        scaled_mu = torch.sum(mu_delta.pow(2.0) / cov_1)
        log_det = torch.sum(torch.log(cov_1)) - torch.sum(torch.log(cov_0))

        return 0.5 * (trace_cov - k + scaled_mu + log_det)


class SymmetricSeparableGaussian(SeparableGaussian):
    """
    Symmetric (antithetic) separable Gaussian distribution
    as used by PGPE.
    """

    MANDATORY_PARAMETERS = {"mu", "sigma"}
    OPTIONAL_PARAMETERS = {"divide_mu_grad_by", "divide_sigma_grad_by", "parenthood_ratio"}

    def _fill(self, out: torch.Tensor, *, generator: Optional[torch.Generator] = None):
        self.make_gaussian(out=out, center=self.mu, stdev=self.sigma, symmetric=True, generator=generator)

    def _compute_gradients(
        self,
        samples: torch.Tensor,
        weights: torch.Tensor,
        ranking_used: Optional[str],
    ) -> dict:
        if "parenthood_ratio" in self.parameters:
            return self._compute_gradients_via_parenthood_ratio(samples, weights)
        else:
            mu = self.mu
            sigma = self.sigma

            # Make sure that the weights (utilities) are 0-centered
            # (Otherwise the formulations would have to consider a bias term)
            if ranking_used not in ("centered", "normalized"):
                weights = weights - torch.mean(weights)

            [nslns] = weights.shape
            # ndirs = nslns // 2

            # Compute the scaled noises, that is, the noise vectors which
            # were used for generating the solutions
            # (solution = scaled_noise + center)
            scaled_noises = samples[0::2] - mu

            # Separate the plus and the minus ends of the directions
            fdplus = weights[0::2]
            fdminus = weights[1::2]

            # Considering that the population is stored like this:
            #                                     _
            #   solution0: center + scaled_noise0  \
            #                                       > direction0
            #   solution1: center - scaled_noise0 _/
            #                                     _
            #   solution2: center + scaled_noise1  \
            #                                       > direction1
            #   solution3: center - scaled_noise1 _/
            #
            #   ...

            # fdplus[0] becomes the utility of the plus end of direction0
            #                   (i.e. utility of solution0)

            # fdminus[0] becomes the utility of the minus end of direction0
            #                   (i.e. utility of solution1)

            # fdplus[1] becomes the utility of the plus end of direction1
            #                   (i.e. utility of solution2)

            # fdminus[1] becomes the utility of the minus end of direction1
            #                   (i.e. utility of solution3)

            # ... and so on...

            grad_mu = self._divide_grad("mu", total(dot((fdplus - fdminus) / 2, scaled_noises)), weights)
            grad_sigma = self._divide_grad(
                "sigma",
                total(dot(((fdplus + fdminus) / 2), ((scaled_noises**2) - (sigma**2)) / sigma)),
                weights,
            )

            return {
                "mu": grad_mu,
                "sigma": grad_sigma,
            }


class ExpSeparableGaussian(SeparableGaussian):
    """exponentialseparable Multivariate Gaussian, as used by SNES"""

    MANDATORY_PARAMETERS = {"mu", "sigma"}
    OPTIONAL_PARAMETERS = set()

    def _compute_gradients(self, samples: torch.Tensor, weights: torch.Tensor, ranking_used: Optional[str]) -> dict:
        if ranking_used != "nes":
            weights = weights / torch.sum(torch.abs(weights))

        scaled_noises = samples - self.mu
        raw_noises = scaled_noises / self.sigma

        mu_grad = total(dot(weights, scaled_noises))
        sigma_grad = total(dot(weights, (raw_noises**2) - 1))

        return {"mu": mu_grad, "sigma": sigma_grad}

    def update_parameters(
        self,
        gradients: dict,
        *,
        learning_rates: Optional[dict] = None,
        optimizers: Optional[dict] = None,
    ) -> "ExpSeparableGaussian":
        mu_grad = gradients["mu"]
        sigma_grad = gradients["sigma"]

        new_mu = self.mu + self._follow_gradient("mu", mu_grad, learning_rates=learning_rates, optimizers=optimizers)
        new_sigma = self.sigma * torch.exp(
            0.5 * self._follow_gradient("sigma", sigma_grad, learning_rates=learning_rates, optimizers=optimizers)
        )

        return self.modified_copy(mu=new_mu, sigma=new_sigma)


class ExpGaussian(Distribution):
    """exponential Multivariate Gaussian, as used by XNES"""

    # Corresponding to mu and A in symbols used in xNES paper
    MANDATORY_PARAMETERS = {"mu", "sigma"}

    # Inverse of sigma, numerically more stable to track this independently to sigma
    OPTIONAL_PARAMETERS = {"sigma_inv"}

    def __init__(
        self,
        parameters: dict,
        *,
        solution_length: Optional[int] = None,
        device: Optional[Device] = None,
        dtype: Optional[DType] = None,
    ):
        [mu_length] = parameters["mu"].shape

        # Make sigma 2D
        if len(parameters["sigma"].shape) == 1:
            parameters["sigma"] = torch.diag(parameters["sigma"])

        # Automatically generate sigma_inv if not provided
        if "sigma_inv" not in parameters:
            parameters["sigma_inv"] = torch.inverse(parameters["sigma"])

        [sigma_length, _] = parameters["sigma"].shape

        if solution_length is None:
            solution_length = mu_length
        else:
            if solution_length != mu_length:
                raise ValueError(
                    f"The argument `solution_length` does not match the length of `mu` provided in `parameters`."
                    f" solution_length={solution_length},"
                    f' parameters["mu"]={mu_length}.'
                )

        if mu_length != sigma_length:
            raise ValueError(
                f"The tensors `mu` and `sigma` provided within `parameters` have mismatching lengths."
                f' parameters["mu"]={mu_length},'
                f' parameters["sigma"]={sigma_length}.'
            )

        super().__init__(
            solution_length=solution_length,
            parameters=parameters,
            device=device,
            dtype=dtype,
        )
        # Make identity matrix as this is used throughout in gradient computation
        self.eye = self.make_zeros((solution_length, solution_length))
        self.eye[range(self.solution_length), range(self.solution_length)] = 1.0

    @property
    def mu(self) -> torch.Tensor:
        """Getter for mu
        Returns:
            mu (torch.Tensor): The center of the search distribution
        """
        return self.parameters["mu"]

    @mu.setter
    def mu(self, new_mu: Iterable):
        """Setter for mu
        Args:
            new_mu (torch.Tensor): The new value of mu
        """
        self.parameters["mu"] = torch.as_tensor(new_mu, dtype=self.dtype, device=self.device)

    @property
    def cov(self) -> torch.Tensor:
        """The covariance matrix A^T A"""
        return self.sigma.transpose(0, 1) @ self.sigma

    @property
    def sigma(self) -> torch.Tensor:
        """Getter for sigma
        Returns:
            sigma (torch.Tensor): The square root of the covariance matrix
        """
        return self.parameters["sigma"]

    @property
    def sigma_inv(self) -> torch.Tensor:
        """Getter for sigma_inv
        Returns:
            sigma_inv (torch.Tensor): The inverse square root of the covariance matrix
        """
        if "sigma_inv" in self.parameters:
            return self.parameters["sigma_inv"]
        else:
            return torch.inverse(self.parameters["sigma"])

    @property
    def A(self) -> torch.Tensor:
        """Alias for self.sigma, for notational consistency with paper"""
        return self.sigma

    @property
    def A_inv(self) -> torch.Tensor:
        """Alias for self.sigma_inv, for notational consistency with paper"""
        return self.sigma_inv

    @sigma.setter
    def sigma(self, new_sigma: Iterable):
        """Setter for sigma
        Args:
            new_sigma (torch.Tensor): The new value of sigma, the square root of the covariance matrix
        """
        self.parameters["sigma"] = torch.as_tensor(new_sigma, dtype=self.dtype, device=self.device)

    def to_global_coordinates(self, local_coordinates: torch.Tensor) -> torch.Tensor:
        """Map samples from local coordinate space N(0, I_d) to global coordinate space N(mu, A^T A)
        This function is the inverse of to_local_coordinates
        Args:
            local_coordinates (torch.Tensor): The local coordinates sampled from N(0, I_d)
        Returns:
            global_coordinates (torch.Tensor): The global coordinates sampled from N(mu, A^T A)
        """
        # Global samples are constructed as x = mu + A z where z is local coordinate
        # We use transpose here to simplify the batched application of A
        return self.mu.unsqueeze(0) + (self.A @ local_coordinates.T).T

    def to_local_coordinates(self, global_coordinates: torch.Tensor) -> torch.Tensor:
        """Map samples from global coordinate space N(mu, A^T A) to local coordinate space N(0, I_d)
        This function is the inverse of to_global_coordinates
        Args:
            global_coordinates (torch.Tensor): The global coordinates sampled from N(mu, A^T A)
        Returns:
            local_coordinates (torch.Tensor): The local coordinates sampled from N(0, I_d)
        """
        # Global samples are constructed as x = mu + A z where z is local coordinate
        # Therefore, we can recover z according to z = A_inv (x - mu)
        return (self.A_inv @ (global_coordinates - self.mu.unsqueeze(0)).T).T

    def _fill(self, out: torch.Tensor, *, generator: Optional[torch.Generator] = None):
        """Fill a tensor with samples from N(mu, A^T A)
        Args:
            out (torch.Tensor): The tensor to fill
            generator (Optional[torch.Generator]): A generator to use to generate random values
        """
        # Fill with local coordinates from N(0, I_d)
        self.make_gaussian(out=out, generator=generator)
        # Map local coordinates to global coordinate system
        out[:] = self.to_global_coordinates(out)

    def _compute_gradients(self, samples: torch.Tensor, weights: torch.Tensor, ranking_used: Optional[str]) -> dict:
        """Compute the gradients with respect to a given set of samples and weights
        Args:
            samples (torch.Tensor): Samples drawn from N(mu, A^T A), ideally using self._fill
            weights (torch.Tensor): Weights e.g. fitnesses or utilities assigned to samples
            ranking_used (optional[str]): The ranking method used to compute weights
        Returns:
            grads (dict): A dictionary containing the approximated natural gradient on d and M
        """
        # Compute the local coordinates
        local_coordinates = self.to_local_coordinates(samples)

        # Make sure that the weights (utilities) are 0-centered
        # (Otherwise the formulations would have to consider a bias term)
        if ranking_used not in ("centered", "normalized"):
            weights = weights - torch.mean(weights)

        d_grad = total(dot(weights, local_coordinates))
        local_coordinates_outer = local_coordinates.unsqueeze(1) * local_coordinates.unsqueeze(2)
        M_grad = torch.sum(
            weights.unsqueeze(-1).unsqueeze(-1) * (local_coordinates_outer - self.eye.unsqueeze(0)), dim=0
        )

        return {
            "d": d_grad,
            "M": M_grad,
        }

    def update_parameters(
        self,
        gradients: dict,
        *,
        learning_rates: Optional[dict] = None,
        optimizers: Optional[dict] = None,
    ) -> "ExpGaussian":
        d_grad = gradients["d"]
        M_grad = gradients["M"]

        if "d" not in learning_rates:
            learning_rates["d"] = learning_rates["mu"]
        if "M" not in learning_rates:
            learning_rates["M"] = learning_rates["sigma"]

        # Follow gradients for d, and M
        update_d = self._follow_gradient("d", d_grad, learning_rates=learning_rates, optimizers=optimizers)
        update_M = self._follow_gradient("M", M_grad, learning_rates=learning_rates, optimizers=optimizers)

        # Fold into parameters mu, A and A inv
        new_mu = self.mu + torch.mv(self.A, update_d)
        new_A = self.A @ torch.matrix_exp(0.5 * update_M)
        new_A_inv = torch.matrix_exp(-0.5 * update_M) @ self.A_inv

        # Return modified distribution
        return self.modified_copy(mu=new_mu, sigma=new_A, sigma_inv=new_A_inv)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides various common operators
to be used within evolutionary algorithms.

Each operator is provided as a separate class,
which is to be instantiated in this form:

    op = OperatorName(
        problem,  # where problem is a Problem instance
        hyperparameter1=...,
        hyperparameter2=...,
        ...
    )

Each operator has its `__call__(...)` method overriden
so that it can be used like a function.
For example, if the operator `op` instantiated above
were a mutation operator, it would be used like this:

    # Apply mutation on a SolutionBatch
    mutated_solution = op(my_solution_batch)

Please see the documentations of the provided operator
classes for details about how to instantiate them,
and how to call them.

A common usage for the operators provided here is to
use them with a genetic algorithm.
More specifically, the SteadyStateGA algorithm provided
within the namespace `evotorch.algorithms` needs
to be configured so that it knows which cross-over operator
and which mutation operator it should apply on the
solutions. The way this is done is as follows:

    import evotorch.algorithms as dra
    import evotorch.operators as dro

    problem = ...   # initialize the Problem

    ga = dra.SteadyStateGA(problem, popsize=...)

    # Configure the genetic algorithm to use
    # simulated binary cross-over
    ga.use(
        dro.SimulatedBinaryCrossOver(
            problem,
            tournament_size=...,
            cross_over_rate=...,
            eta=...
        )
    )

    # Configure the genetic algorithm to use
    # Gaussian mutation
    ga.use(
        dro.GaussianMutation(
            problem,
            stdev=...
        )
    )
"""

__all__ = (
    "CopyingOperator",
    "CrossOver",
    "Operator",
    "SingleObjOperator",
    "CosynePermutation",
    "GaussianMutation",
    "OnePointCrossOver",
    "SimulatedBinaryCrossOver",
    "CutAndSplice",
)


from . import base, real, sequence
from .base import CopyingOperator, CrossOver, Operator, SingleObjOperator
from .real import CosynePermutation, GaussianMutation, OnePointCrossOver, SimulatedBinaryCrossOver
from .sequence import CutAndSplice



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Hook class, which is used for event handling,
and for defining additional behaviors to the class instances which own
the Hook.
"""

from collections.abc import Mapping, MutableSequence
from typing import Any, Callable, Iterable, Optional, Union


class Hook(MutableSequence):
    """
    A Hook stores a list of callable objects to be called for handling
    certain events. A Hook itself is callable, which invokes the callables
    stored in its list. If the callables stored by the Hook return list-like
    objects or dict-like objects, their returned results are accumulated,
    and then those accumulated results are finally returned by the Hook.
    """

    def __init__(
        self,
        callables: Optional[Iterable[Callable]] = None,
        *,
        args: Optional[Iterable] = None,
        kwargs: Optional[Mapping] = None,
    ):
        """
        Initialize the Hook.

        Args:
            callables: A sequence of callables to be stored by the Hook.
            args: Positional arguments which, when the Hook is called,
                are to be passed to every callable stored by the Hook.
                Please note that these positional arguments will be passed
                as the leftmost arguments, and, the other positional
                arguments passed via the `__call__(...)` method of the
                Hook will be added to the right of these arguments.
            kwargs: Keyword arguments which, when the Hook is called,
                are to be passed to every callable stored by the Hook.
                Please note that these keyword arguments could be overriden
                by the keyword arguments passed via the `__call__(...)`
                method of the Hook.
        """
        self._funcs: list = [] if callables is None else list(callables)
        self._args: list = [] if args is None else list(args)
        self._kwargs: dict = {} if kwargs is None else dict(kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Optional[Union[dict, list]]:
        """
        Call every callable object stored by the Hook.
        The results of the stored callable objects (which can be dict-like
        or list-like objects) are accumulated and finally returned.

        Args:
            args: Additional positional arguments to be passed to the stored
                callables.
            kwargs: Additional keyword arguments to be passed to the stored
                keyword arguments.
        """

        all_args = []
        all_args.extend(self._args)
        all_args.extend(args)

        all_kwargs = {}
        all_kwargs.update(self._kwargs)
        all_kwargs.update(kwargs)

        result: Optional[Union[dict, list]] = None

        for f in self._funcs:
            tmp = f(*all_args, **all_kwargs)
            if tmp is not None:
                if isinstance(tmp, Mapping):
                    if result is None:
                        result = dict(tmp)
                    elif isinstance(result, list):
                        raise TypeError(
                            f"The function {f} returned a dict-like object."
                            f" However, previous function(s) in this hook had returned list-like object(s)."
                            f" Such incompatible results cannot be accumulated."
                        )
                    elif isinstance(result, dict):
                        result.update(tmp)
                    else:
                        raise RuntimeError
                elif isinstance(tmp, Iterable):
                    if result is None:
                        result = list(tmp)
                    elif isinstance(result, list):
                        result.extend(tmp)
                    elif isinstance(result, dict):
                        raise TypeError(
                            f"The function {f} returned a list-like object."
                            f" However, previous function(s) in this hook had returned dict-like object(s)."
                            f" Such incompatible results cannot be accumulated."
                        )
                    else:
                        raise RuntimeError
                else:
                    raise TypeError(
                        f"Expected the function {f} to return None, or a dict-like object, or a list-like object."
                        f" However, the function returned an object of type {repr(type(tmp))}."
                    )

        return result

    def accumulate_dict(self, *args: Any, **kwargs: Any) -> Optional[Union[dict, list]]:
        result = self(*args, **kwargs)
        if result is None:
            return {}
        elif isinstance(result, Mapping):
            return result
        else:
            raise TypeError(
                f"Expected the functions in this hook to accumulate"
                f" dictionary-like objects. Instead, accumulated"
                f" an object of type {type(result)}."
                f" Hint: are the functions registered in this hook"
                f" returning non-dictionary iterables?"
            )

    def accumulate_sequence(self, *args: Any, **kwargs: Any) -> Optional[Union[dict, list]]:
        result = self(*args, **kwargs)
        if result is None:
            return []
        elif isinstance(result, Mapping):
            raise TypeError(
                f"Expected the functions in this hook to accumulate"
                f" sequences (that are NOT dictionaries). Instead, accumulated"
                f" a dict-like object of type {type(result)}."
                f" Hint: are the functions registered in this hook"
                f" returning objects with Mapping interface?"
            )
        else:
            return result

    def _to_string(self) -> str:
        init_args = [repr(self._funcs)]

        if len(self._args) > 0:
            init_args.append(f"args={self._args}")

        if len(self._kwargs) > 0:
            init_args.append(f"kwargs={self._kwargs}")

        s_init_args = ", ".join(init_args)

        return f"{type(self).__name__}({s_init_args})"

    def __repr__(self) -> str:
        return self._to_string()

    def __str__(self) -> str:
        return self._to_string()

    def __getitem__(self, i: Union[int, slice]) -> Union[Callable, "Hook"]:
        if isinstance(i, slice):
            return Hook(self._funcs[i], args=self._args, kwargs=self._kwargs)
        else:
            return self._funcs[i]

    def __setitem__(self, i: Union[int, slice], x: Iterable[Callable]):
        self._funcs[i] = x

    def __delitem__(self, i: Union[int, slice]):
        del self._funcs[i]

    def insert(self, i: int, x: Callable):
        self._funcs.insert(i, x)

    def __len__(self) -> int:
        return len(self._funcs)

    @property
    def args(self) -> list:
        """Positional arguments that will be passed to the stored callables"""
        return self._args

    @property
    def kwargs(self) -> dict:
        """Keyword arguments that will be passed to the stored callables"""
        return self._kwargs
